"""Pulse and PulseSequence classes."""
import re
from abc import ABC, abstractmethod
from enum import Enum

import numpy as np
from qibo.config import log
from scipy.signal import lfilter

from qibolab.symbolic import intSymbolicExpression as se_int


class PulseType(Enum):
    """An enumeration to distinguish different types of pulses.

    READOUT pulses triger acquisitions.
    DRIVE pulses are used to control qubit states.
    FLUX pulses are used to shift the frequency of flux tunable qubits and with it implement two-qubit gates.
    """

    READOUT = "ro"
    DRIVE = "qd"
    FLUX = "qf"


class Waveform:
    """A class to save pulse waveforms.

    A waveform is a list of samples, or discrete data points, used by the digital to analogue converters (DACs)
    to synthesise pulses.

    Attributes:
        data (np.ndarray): a numpy array containing the samples.
        serial (str): a string that can be used as a lable to identify the waveform. It is not automatically
            generated, it must be set by the user.
    """

    DECIMALS = 5

    def __init__(self, data):
        """Initialises the waveform with a of samples."""

        self.data: np.ndarray = np.array(data)
        self.serial: str = ""

    def __len__(self):
        """Returns the length of the waveform, the number of samples."""

        return len(self.data)

    def __eq__(self, other):
        """Compares two waveforms.

        Two waveforms are considered equal if their samples, rounded to `Waveform.DECIMALS` decimal places,
        are all equal.
        """

        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """Returns a hash of the array of data, after rounding each sample to `Waveform.DECIMALS` decimal places."""

        return hash(str(np.around(self.data, Waveform.DECIMALS) + 0))

    def __repr__(self):
        """Returns the waveform serial as its string representation."""

        return self.serial

    def plot(self, savefig_filename=None):
        """Plots the waveform.

        Args:
            savefig_filename (str): a file path. If provided the plot is save to a file.
        """

        import matplotlib.pyplot as plt

        plt.figure(figsize=(14, 5), dpi=200)
        plt.plot(self.data, c="C0", linestyle="dashed")
        plt.xlabel("Sample Number")
        plt.ylabel("Amplitude")
        plt.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
        plt.suptitle(self.serial)
        if savefig_filename:
            plt.savefig(savefig_filename)
        else:
            plt.show()
        plt.close()


class ShapeInitError(RuntimeError):
    """Error raised when a pulse has not been fully defined."""

    default_msg = "PulseShape attribute pulse must be initialised in order to be able to generate pulse waveforms"

    def __init__(self, msg=None, *args):
        if msg is None:
            msg = self.default_msg
        super().__init__(msg, *args)


class PulseShape(ABC):
    """Abstract class for pulse shapes.

    A PulseShape object is responsible for generating envelope and modulated waveforms from a set
    of pulse parameters, its type and a predefined SAMPLING_RATE. PulsShape generates both i (in-phase)
    and q (quadrature) components.
    """

    SAMPLING_RATE = 1e9  # 1GSaPS
    """SAMPLING_RATE (int): sampling rate in samples per second (SaPS)"""
    pulse = None
    """pulse (Pulse): the pulse associated with it. Its parameters are used to generate pulse waveforms."""

    @property
    @abstractmethod
    def envelope_waveform_i(self) -> Waveform:  # pragma: no cover
        raise NotImplementedError

    @property
    @abstractmethod
    def envelope_waveform_q(self) -> Waveform:  # pragma: no cover
        raise NotImplementedError

    @property
    def envelope_waveforms(self):  #  -> tuple[Waveform, Waveform]:  # pragma: no cover
        """A tuple with the i and q envelope waveforms of the pulse."""

        return (self.envelope_waveform_i, self.envelope_waveform_q)

    @property
    def modulated_waveform_i(self) -> Waveform:
        """The waveform of the i component of the pulse, modulated with its frequency."""

        return self.modulated_waveforms[0]

    @property
    def modulated_waveform_q(self) -> Waveform:
        """The waveform of the q component of the pulse, modulated with its frequency."""

        return self.modulated_waveforms[1]

    @property
    def modulated_waveforms(self):
        """A tuple with the i and q waveforms of the pulse, modulated with its frequency."""

        if not self.pulse:
            raise ShapeInitError

        pulse = self.pulse
        if abs(pulse._if) * 2 > PulseShape.SAMPLING_RATE:
            log.info(
                f"WARNING: The frequency of pulse {pulse.serial} is higher than the nyqusit frequency ({int(PulseShape.SAMPLING_RATE // 2)}) for the device sampling rate: {int(PulseShape.SAMPLING_RATE)}"
            )
        num_samples = int(np.rint(pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
        time = np.arange(num_samples) / PulseShape.SAMPLING_RATE
        global_phase = pulse.global_phase
        cosalpha = np.cos(2 * np.pi * pulse._if * time + global_phase + pulse.relative_phase)
        sinalpha = np.sin(2 * np.pi * pulse._if * time + global_phase + pulse.relative_phase)

        mod_matrix = np.array([[cosalpha, -sinalpha], [sinalpha, cosalpha]]) / np.sqrt(2)

        (envelope_waveform_i, envelope_waveform_q) = self.envelope_waveforms
        result = []
        for n, t, ii, qq in zip(
            np.arange(num_samples),
            time,
            envelope_waveform_i.data,
            envelope_waveform_q.data,
        ):
            result.append(mod_matrix[:, :, n] @ np.array([ii, qq]))
        mod_signals = np.array(result)

        modulated_waveform_i = Waveform(mod_signals[:, 0])
        modulated_waveform_i.serial = f"Modulated_Waveform_I(num_samples = {num_samples}, amplitude = {format(pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {str(pulse.shape)}, frequency = {format(pulse._if, '_')}, phase = {format(global_phase + pulse.relative_phase, '.6f').rstrip('0').rstrip('.')})"
        modulated_waveform_q = Waveform(mod_signals[:, 1])
        modulated_waveform_q.serial = f"Modulated_Waveform_Q(num_samples = {num_samples}, amplitude = {format(pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {str(pulse.shape)}, frequency = {format(pulse._if, '_')}, phase = {format(global_phase + pulse.relative_phase, '.6f').rstrip('0').rstrip('.')})"
        return (modulated_waveform_i, modulated_waveform_q)

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        return isinstance(item, type(self))


class Rectangular(PulseShape):
    """Rectangular pulse shape."""

    def __init__(self):
        self.name = "Rectangular"
        self.pulse: Pulse = None

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            waveform = Waveform(self.pulse.amplitude * np.ones(num_samples))
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            waveform = Waveform(np.zeros(num_samples))
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}()"


class Exponential(PulseShape):
    r"""Exponential pulse shape (Square pulse with an exponential decay).

    Args:
        tau (float): Parameter that controls the decay of the first exponential function
        upsilon (float): Parameter that controls the decay of the second exponential function
        g (float): Parameter that weights the second exponential function


    .. math::

        A\frac{\exp\left(-\frac{x}{\text{upsilon}}\right) + g \exp\left(-\frac{x}{\text{tau}}\right)}{1 + g}

    """

    def __init__(self, tau: float, upsilon: float, g: float = 0.1):
        self.name = "Exponential"
        self.pulse: Pulse = None
        self.tau: float = float(tau)
        self.upsilon: float = float(upsilon)
        self.g: float = float(g)

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            x = np.arange(0, num_samples, 1)
            waveform = Waveform(
                self.pulse.amplitude
                * ((np.ones(num_samples) * np.exp(-x / self.upsilon)) + self.g * np.exp(-x / self.tau))
                / (1 + self.g)
            )

            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            waveform = Waveform(np.zeros(num_samples))
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({format(self.tau, '.3f').rstrip('0').rstrip('.')}, {format(self.upsilon, '.3f').rstrip('0').rstrip('.')}, {format(self.g, '.3f').rstrip('0').rstrip('.')})"


class Gaussian(PulseShape):
    r"""Gaussian pulse shape.

    Args:
        rel_sigma (float): relative sigma so that the pulse standard deviation (sigma) = duration / rel_sigma

    .. math::

        A\exp^{-\frac{1}{2}\frac{(t-\mu)^2}{\sigma^2}}
    """

    def __init__(self, rel_sigma: float):
        self.name = "Gaussian"
        self.pulse: Pulse = None
        self.rel_sigma: float = float(rel_sigma)

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        if super().__eq__(item):
            return self.rel_sigma == item.rel_sigma
        return False

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            x = np.arange(0, num_samples, 1)
            waveform = Waveform(
                self.pulse.amplitude
                * np.exp(-(1 / 2) * (((x - (num_samples - 1) / 2) ** 2) / (((num_samples) / self.rel_sigma) ** 2)))
            )
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            waveform = Waveform(np.zeros(num_samples))
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({format(self.rel_sigma, '.6f').rstrip('0').rstrip('.')})"


class Drag(PulseShape):
    """
    Derivative Removal by Adiabatic Gate (DRAG) pulse shape.

    Args:
        rel_sigma (float): relative sigma so that the pulse standard deviation (sigma) = duration / rel_sigma
        beta (float): relative sigma so that the pulse standard deviation (sigma) = duration / rel_sigma
    .. math::


    """

    def __init__(self, rel_sigma, beta):
        self.name = "Drag"
        self.pulse: Pulse = None
        self.rel_sigma = float(rel_sigma)
        self.beta = float(beta)

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        if super().__eq__(item):
            return self.rel_sigma == item.rel_sigma and self.beta == item.beta
        return False

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            x = np.arange(0, num_samples, 1)
            i = self.pulse.amplitude * np.exp(
                -(1 / 2) * (((x - (num_samples - 1) / 2) ** 2) / (((num_samples) / self.rel_sigma) ** 2))
            )
            waveform = Waveform(i)
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            x = np.arange(0, num_samples, 1)
            i = self.pulse.amplitude * np.exp(
                -(1 / 2) * (((x - (num_samples - 1) / 2) ** 2) / (((num_samples) / self.rel_sigma) ** 2))
            )
            q = (
                self.beta
                * (-(x - (num_samples - 1) / 2) / ((num_samples / self.rel_sigma) ** 2))
                * i
                * PulseShape.SAMPLING_RATE
                / 1e9
            )
            waveform = Waveform(q)
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({format(self.rel_sigma, '.6f').rstrip('0').rstrip('.')}, {format(self.beta, '.6f').rstrip('0').rstrip('.')})"


class IIR(PulseShape):
    """IIR Filter using scipy.signal lfilter."""

    # https://arxiv.org/pdf/1907.04818.pdf (page 11 - filter formula S22)
    # p = [A, tau_iir]
    # p = [b0 = 1−k +k ·α, b1 = −(1−k)·(1−α),a0 = 1 and a1 = −(1−α)]
    # p = [b0, b1, a0, a1]

    def __init__(self, b, a, target: PulseShape):
        self.name = "IIR"
        self.target: PulseShape = target
        self._pulse: Pulse = None
        self.a: np.ndarray = np.array(a)
        self.b: np.ndarray = np.array(b)
        # Check len(a) = len(b) = 2

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        if super().__eq__(item):
            return self.target == item.target and (self.a == item.a).all() and (self.b == item.b).all()
        return False

    @property
    def pulse(self):
        return self._pulse

    @pulse.setter
    def pulse(self, value):
        self._pulse = value
        self.target.pulse = value

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            self.a = self.a / self.a[0]
            gain = np.sum(self.b) / np.sum(self.a)
            if not gain == 0:
                self.b = self.b / gain
            data = lfilter(b=self.b, a=self.a, x=self.target.envelope_waveform_i.data)
            if not np.max(np.abs(data)) == 0:
                data = data / np.max(np.abs(data))
            data = np.abs(self.pulse.amplitude) * data
            waveform = Waveform(data)
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            self.a = self.a / self.a[0]
            gain = np.sum(self.b) / np.sum(self.a)
            if not gain == 0:
                self.b = self.b / gain
            data = lfilter(b=self.b, a=self.a, x=self.target.envelope_waveform_q.data)
            if not np.max(np.abs(data)) == 0:
                data = data / np.max(np.abs(data))
            data = np.abs(self.pulse.amplitude) * data
            waveform = Waveform(data)
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        formatted_b = [round(b, 3) for b in self.b]
        formatted_a = [round(a, 3) for a in self.a]
        return f"{self.name}({formatted_b}, {formatted_a}, {self.target})"


class SNZ(PulseShape):
    """
    Sudden variant Net Zero.
    https://arxiv.org/abs/2008.07411

    """

    def __init__(self, t_half_flux_pulse=None, b_amplitude=1):
        self.name = "SNZ"
        self.pulse: Pulse = None
        self.t_half_flux_pulse: float = t_half_flux_pulse
        self.b_amplitude: float = b_amplitude

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        if super().__eq__(item):
            return self.t_half_flux_pulse == item.t_half_flux_pulse and self.b_amplitude == item.b_amplitude
        return False

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            if not self.t_half_flux_pulse:
                self.t_half_flux_pulse = self.pulse.duration / 2
            elif 2 * self.t_half_flux_pulse > self.pulse.duration:
                raise ValueError("Pulse shape parameter error: pulse.t_half_flux_pulse <= pulse.duration")
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            half_flux_pulse_samples = int(np.rint(num_samples * self.t_half_flux_pulse / self.pulse.duration))
            iding_samples = num_samples - 2 * half_flux_pulse_samples
            waveform = Waveform(
                np.concatenate(
                    (
                        self.pulse.amplitude * np.ones(half_flux_pulse_samples - 1),
                        np.array([self.pulse.amplitude * self.b_amplitude]),
                        np.zeros(iding_samples),
                        -1 * np.array([self.pulse.amplitude * self.b_amplitude]),
                        -1 * self.pulse.amplitude * np.ones(half_flux_pulse_samples - 1),
                    )
                )
            )
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))
            waveform = Waveform(np.zeros(num_samples))
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({self.t_half_flux_pulse}, {self.b_amplitude})"


class eCap(PulseShape):
    r"""eCap pulse shape.

    Args:
        alpha (float):

    .. math::

        e_{\cap(t,\alpha)} &=& A[1 + \tanh(\alpha t/t_\theta)][1 + \tanh(\alpha (1 - t/t_\theta))]\\
        &\times& [1 + \tanh(\alpha/2)]^{-2}

    """

    def __init__(self, alpha: float):
        self.name = "eCap"
        self.pulse: Pulse = None
        self.alpha: float = float(alpha)

    def __eq__(self, item) -> bool:
        """Overloads == operator"""
        if super().__eq__(item):
            return self.alpha == item.alpha
        return False

    @property
    def envelope_waveform_i(self) -> Waveform:
        if self.pulse:
            num_samples = int(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE)
            x = np.arange(0, num_samples, 1)
            waveform = Waveform(
                self.pulse.amplitude
                * (1 + np.tanh(self.alpha * x / num_samples))
                * (1 + np.tanh(self.alpha * (1 - x / num_samples)))
                / (1 + np.tanh(self.alpha / 2)) ** 2
            )
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        if self.pulse:
            num_samples = int(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE)
            waveform = Waveform(np.zeros(num_samples))
            waveform.serial = f"Envelope_Waveform_Q(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({format(self.alpha, '.6f').rstrip('0').rstrip('.')})"


class Custom(PulseShape):
    """Arbitrary shape."""

    def __init__(self, envelope):
        self.name = "Custom"
        self._pulse: Pulse = None
        self.envelope: np.ndarray = np.array(envelope)

    @property
    def pulse(self):
        return self._pulse

    @pulse.setter
    def pulse(self, value):
        self._pulse = value

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        if self.pulse:
            assert self.pulse.duration == len(self.envelope)
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))

            waveform = Waveform(self.envelope * self.pulse.amplitude)
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        if self.pulse:
            assert self.pulse.duration == len(self.envelope)
            num_samples = int(np.rint(self.pulse.duration / 1e9 * PulseShape.SAMPLING_RATE))

            waveform = Waveform(self.envelope * self.pulse.amplitude)
            waveform.serial = f"Envelope_Waveform_I(num_samples = {num_samples}, amplitude = {format(self.pulse.amplitude, '.6f').rstrip('0').rstrip('.')}, shape = {repr(self)})"
            return waveform
        raise ShapeInitError

    def __repr__(self):
        return f"{self.name}({self.envelope[:3]}, ...)"


class Pulse:
    """A class to represent a pulse to be sent to the QPU.

    Args:
        start (int | intSymbolicExpression): Start time of pulse in ns.
        duration (int | intSymbolicExpression): Pulse duration in ns.
        amplitude (float): Pulse digital amplitude (unitless) [-1 to 1].
        frequency (int): Pulse Intermediate Frequency in Hz [10e6 to 300e6].
        relative_phase (float): To be added.
        shape: (PulseShape | str): {'Rectangular()', 'Gaussian(rel_sigma)', 'DRAG(rel_sigma, beta)'} Pulse shape.
            See :py:mod:`qibolab.pulses` for list of available shapes.
        channel (int | str): the channel on which the pulse should be synthesised.
        type (PulseType | str): {'ro', 'qd', 'qf'} type of pulse {ReadOut, Qubit Drive, Qubit Flux}
        qubit (int): qubit associated with the pulse

    Example:
        .. code-block:: python

            from qibolab.pulses import Pulse, Gaussian

            # define Gaussian drive pulse
            drive_pulse = Pulse(
                start=0,
                duration=60,
                amplitude=0.3,
                frequency=-200_000_000,
                relative_phase=0.0,
                shape=Gaussian(5),
                channel=1,
                type=PulseType.DRIVE,
                qubit=0,
            )

            # define Rectangular readout pulse
            readout_pulse = Pulse(
                start=intSymbolicExpression(60),
                duration=2000,
                amplitude=0.3,
                frequency=20_000_000,
                relative_phase=0.0,
                shape=Rectangular(),
                channel=2,
                type=PulseType.READOUT,
                qubit=0,
            )
    """

    count: int = 0

    def __init__(
        self,
        start,
        duration,
        amplitude,
        frequency,
        relative_phase,
        shape,
        channel,
        type=PulseType.DRIVE,
        qubit=0,
    ):
        # def __init__(self, start:int | se_int, duration:int | se_int, amplitude:float, frequency:int, relative_phase:float, shape: PulseShape | str,
        #                    channel: int | str, type: PulseType | str  = PulseType.DRIVE, qubit: int | str = 0):

        self._start: se_int = None
        self._duration: se_int = None
        self._finish: se_int = None
        self._amplitude: float = None
        self._frequency: int = None
        self._relative_phase: float = None
        self._shape: PulseShape = None
        self._channel = None
        # self._channel: int | str = None
        self._type: PulseType = None
        self._qubit = None
        # self._qubit: int | str = None
        self._id: int = Pulse.count
        Pulse.count += 1

        self.start = start
        self.duration = duration
        self.amplitude = amplitude
        self.frequency = frequency
        self.relative_phase = relative_phase
        self.shape = shape
        self.channel = channel
        self.type = type
        self.qubit = qubit

        self._if = 0

    def __del__(self):
        del self._start
        del self._duration
        del self._finish
        # del self._shape TODO activate when returning a deep copy of shape or when making a deep copy of shape in init

    @property
    def start(self) -> int:
        """Returns the time when the pulse is scheduled to be played, in ns."""
        if isinstance(self._start, se_int):
            return self._start.value
        return self._start

    @start.setter
    def start(self, value):
        """Sets the time when the pulse is scheduled to be played.

        Args:
            value (se_int | int | np.integer): the time in ns.
        """

        if not isinstance(value, (se_int, int, np.integer)):
            raise TypeError(f"start argument type should be intSymbolicExpression or int, got {type(value).__name__}")
        if not value >= 0:
            raise ValueError(f"start argument must be >= 0, got {value}")

        if isinstance(value, se_int):
            self._start = se_int(value.symbol)["_p" + str(self._id) + "_start"]

        elif isinstance(self._start, se_int):
            if isinstance(value, np.integer):
                self._start.value = int(value)
            elif isinstance(value, int):
                self._start.value = value
        else:
            if isinstance(value, np.integer):
                self._start = int(value)
            elif isinstance(value, int):
                self._start = value

        if not self._duration is None:
            if (
                isinstance(self._start, se_int)
                or isinstance(self._duration, se_int)
                or isinstance(self._finish, se_int)
            ):
                self._finish = se_int(self._start + self._duration)["_p" + str(self._id) + "_finish"]
            else:
                self._finish = self._start + self._duration

    @property
    def duration(self) -> int:
        """Returns the duration of the pulse, in ns."""
        if isinstance(self._duration, se_int):
            return self._duration.value
        return self._duration

    @duration.setter
    def duration(self, value):
        """Sets the duration of the pulse.

        Args:
            value (se_int | int | np.integer): the time in ns.
        """

        if not isinstance(value, (se_int, int, np.integer)):
            raise TypeError(
                f"duration argument type should be intSymbolicExpression or int, got {type(value).__name__}"
            )
        if not value >= 0:
            raise ValueError(f"duration argument must be >= 0, got {value}")
        if isinstance(value, se_int):
            self._duration = se_int(value.symbol)["_p" + str(self._id) + "_duration"]

        elif isinstance(self._duration, se_int):
            if isinstance(value, np.integer):
                self._duration.value = int(value)
            elif isinstance(value, int):
                self._duration.value = value
        else:
            if isinstance(value, np.integer):
                self._duration = int(value)
            elif isinstance(value, int):
                self._duration = value

        if not self._start is None:
            if (
                isinstance(self._start, se_int)
                or isinstance(self._duration, se_int)
                or isinstance(self._finish, se_int)
            ):
                self._finish = se_int(self._start + self._duration)["_p" + str(self._id) + "_finish"]
            else:
                self._finish = self._start + self._duration

    @property
    def finish(self) -> int:
        """Returns the time when the pulse is scheduled to finish.

        Calculated as pulse.start - pulse finish.
        """
        if isinstance(self._finish, se_int):
            return self._finish.value
        return self._finish

    @property
    def se_start(self) -> se_int:
        """Returns a symbolic expression for the pulse start."""

        if not isinstance(self._start, se_int):
            self._start = se_int(self._start)["_p" + str(self._id) + "_start"]
        return self._start

    @property
    def se_duration(self) -> se_int:
        """Returns a symbolic expression for the pulse duration."""

        if not isinstance(self._duration, se_int):
            self._duration = se_int(self._duration)["_p" + str(self._id) + "_duration"]
        return self._duration

    @property
    def se_finish(self) -> se_int:
        """Returns a symbolic expression for the pulse finish."""

        if not isinstance(self._finish, se_int):
            self._finish = se_int(self._finish)["_p" + str(self._id) + "_finish"]
        return self._finish

    @property
    def amplitude(self) -> float:
        """Returns the amplitude of the pulse.

        Pulse amplitudes are normalised between -1 and 1.
        """

        return self._amplitude

    @amplitude.setter
    def amplitude(self, value):
        """Sets the amplitude of the pulse.

        Args:
            value (int | float | np.floating): a unitless value between -1 and 1.
        """

        if isinstance(value, int):
            value = float(value)
        if not isinstance(value, (float, np.floating)):
            raise TypeError(f"amplitude argument type should be float, got {type(value).__name__}")
        if not ((value >= -1) & (value <= 1)):
            raise ValueError(f"amplitude argument must be >= -1 & <= 1, got {value}")
        if isinstance(value, np.floating):
            self._amplitude = float(value)
        elif isinstance(value, float):
            self._amplitude = value

    @property
    def frequency(self) -> int:
        """Returns the frequency of the pulse, in Hz."""

        return self._frequency

    @frequency.setter
    def frequency(self, value):
        """Sets the frequency of the pulse.

        Args:
            value (int | float | np.integer | np.floating): the frequency in Hz.
        """

        if not isinstance(value, (int, float, np.integer, np.floating)):
            raise TypeError(f"frequency argument type should be int, got {type(value).__name__}")
        if isinstance(value, (float, np.integer, np.floating)):
            self._frequency = int(value)
        elif isinstance(value, int):
            self._frequency = value

    @property
    def global_phase(self):
        """Returns the global phase of the pulse, in radians.

        This phase is calculated from the pulse start time and frequency as `2 * pi * frequency * start`.
        """

        # pulse start, duration and finish are in ns
        return 2 * np.pi * self._frequency * self.start / 1e9

    @property
    def relative_phase(self) -> float:
        """Returns the relative phase of the pulse, in radians."""

        return self._relative_phase

    @relative_phase.setter
    def relative_phase(self, value):
        """Sets a relative phase for the pulse.

        Args:
            value (int | float | np.integer | np.floating): the relative phase in radians.
        """

        if not isinstance(value, (int, float, np.integer, np.floating)):
            raise TypeError(f"relative_phase argument type should be int or float, got {type(value).__name__}")
        if isinstance(value, (int, np.integer, np.floating)):
            self._relative_phase = float(value)
        elif isinstance(value, float):
            self._relative_phase = value

    @property
    def phase(self) -> float:
        """Returns the total phase of the pulse, in radians.

        The total phase is computed as the sum of the global and relative phases.
        """
        return 2 * np.pi * self._frequency * self.start / 1e9 + self._relative_phase

    @property
    def shape(self) -> PulseShape:
        """Returns the shape of the pulse, as a PulseShape object."""

        return self._shape

    @shape.setter
    def shape(self, value):
        """Sets the shape of the pulse.

        Args:
            value (PulseShape | str): a string representing the pulse shape and its main parameters, or a PulseShape object.
        """

        if not isinstance(value, (PulseShape, str)):
            raise TypeError(f"shape argument type should be PulseShape or str, got {type(value).__name__}")
        if isinstance(value, PulseShape):
            self._shape = value
        elif isinstance(value, str):
            shape_name = re.findall(r"(\w+)", value)[0]
            if shape_name not in globals():
                raise ValueError(f"shape {value} not found")
            shape_parameters = re.findall(r"[\w+\d\.\d]+", value)[1:]
            # TODO: create multiple tests to prove regex working correctly
            self._shape = globals()[shape_name](*shape_parameters)

        # link the pulse attribute of the PulseShape object to the pulse.
        self._shape.pulse = self

    @property
    def channel(self):
        """Returns the channel on which the pulse should be played.

        When a sequence of pulses is sent to the platform for execution, each pulse is sent to the instrument
        responsible for playing pulses the pulse channel. The connection of instruments with channels is defined
        in the platform runcard.
        """

        # def channel(self) -> int | str:
        return self._channel

    @channel.setter
    def channel(self, value):
        """Sets the channel on which the pulse should be played.

        Args:
            value (int | str): an integer or a string used to identify the channel.
        """

        if not isinstance(value, (int, str)):
            raise TypeError(f"channel argument type should be int or str, got {type(value).__name__}")
        self._channel = value

    @property
    def type(self) -> PulseType:
        """Returns the pulse type, as an element of PulseType enumeration."""

        return self._type

    @type.setter
    def type(self, value):
        """Sets the type of the pulse.

        Args:
            value (PulseType | str): the type of pulse as an element of PulseType enumeration or as a two-letter string.
        """

        if isinstance(value, PulseType):
            self._type = value
        elif isinstance(value, str):
            self._type = PulseType(value)
        else:
            raise TypeError(f"type argument should be PulseType or str, got {type(value).__name__}")

    @property
    def qubit(self):
        """Returns the qubit addressed by the pulse."""

        # def qubit(self) -> int | str:
        return self._qubit

    @qubit.setter
    def qubit(self, value):
        """Sets the qubit addressed by the pulse.

        Args:
            value (int | str): an integer or a string used to identify the qubit.
        """

        if not isinstance(value, (int, str)):
            raise TypeError(f"qubit argument type should be int or str, got {type(value).__name__}")
        self._qubit = value

    @property
    def serial(self) -> str:
        """Returns a string representation of the pulse."""

        return f"Pulse({self.start}, {self.duration}, {format(self.amplitude, '.6f').rstrip('0').rstrip('.')}, {format(self.frequency, '_')}, {format(self.relative_phase, '.6f').rstrip('0').rstrip('.')}, {self.shape}, {self.channel}, {self.type}, {self.qubit})"

    @property
    def id(self) -> int:
        return id(self)

    @property
    def envelope_waveform_i(self) -> Waveform:
        """The envelope waveform of the i component of the pulse."""

        return self._shape.envelope_waveform_i

    @property
    def envelope_waveform_q(self) -> Waveform:
        """The envelope waveform of the q component of the pulse."""

        return self._shape.envelope_waveform_q

    @property
    def envelope_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        """A tuple with the i and q envelope waveforms of the pulse."""

        return (self._shape.envelope_waveform_i, self._shape.envelope_waveform_q)

    @property
    def modulated_waveform_i(self) -> Waveform:
        """The waveform of the i component of the pulse, modulated with its frequency."""

        return self._shape.modulated_waveform_i

    @property
    def modulated_waveform_q(self) -> Waveform:
        """The waveform of the q component of the pulse, modulated with its frequency."""

        return self._shape.modulated_waveform_q

    @property
    def modulated_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        """A tuple with the i and q waveforms of the pulse, modulated with its frequency."""

        return self._shape.modulated_waveforms

    def __repr__(self):
        return self.serial

    def __hash__(self):
        return hash(self.serial)

    def __eq__(self, other):
        if isinstance(other, Pulse):
            return self.serial == other.serial
        return False

    def __add__(self, other):
        if isinstance(other, Pulse):
            return PulseSequence(self, other)
        if isinstance(other, PulseSequence):
            return PulseSequence(self, *other.pulses)
        raise TypeError(f"Expected Pulse or PulseSequence; got {type(other).__name__}")

    def __mul__(self, n):
        if not isinstance(n, int):
            raise TypeError(f"Expected int; got {type(n).__name__}")
        if n < 0:
            raise TypeError(f"argument n should be >=0, got {n}")
        return PulseSequence(*([self.copy()] * n))

    def __rmul__(self, n):
        return self.__mul__(n)

    def copy(self):  # -> Pulse|ReadoutPulse|DrivePulse|FluxPulse:
        """Returns a new Pulse object with the same attributes."""

        if type(self) == ReadoutPulse:
            return ReadoutPulse(
                self.start,
                self.duration,
                self.amplitude,
                self.frequency,
                self.relative_phase,
                repr(self._shape),  # self._shape,
                self.channel,
                self.qubit,
            )
        elif type(self) == DrivePulse:
            return DrivePulse(
                self.start,
                self.duration,
                self.amplitude,
                self.frequency,
                self.relative_phase,
                repr(self._shape),  # self._shape,
                self.channel,
                self.qubit,
            )

        elif type(self) == FluxPulse:
            return FluxPulse(
                self.start,
                self.duration,
                self.amplitude,
                self._shape,
                self.channel,
                self.qubit,
            )
        else:
            # return eval(self.serial)
            return Pulse(
                self.start,
                self.duration,
                self.amplitude,
                self.frequency,
                self.relative_phase,
                repr(self._shape),  # self._shape,
                self.channel,
                self.type,
                self.qubit,
            )

    def shallow_copy(self):  # -> Pulse:
        return Pulse(
            self._start,
            self._duration,
            self._amplitude,
            self._frequency,
            self._relative_phase,
            self._shape,
            self._channel,
            self._type,
            self._qubit,
        )

    def is_equal_ignoring_start(self, item) -> bool:
        """Check if two pulses are equal ignoring start time"""
        return (
            self.duration == item.duration
            and self.amplitude == item.amplitude
            and self.frequency == item.frequency
            and self.relative_phase == item.relative_phase
            and self.shape == item.shape
            and self.channel == item.channel
            and self.type == item.type
            and self.qubit == item.qubit
        )

    def plot(self, savefig_filename=None):
        """Plots the pulse envelope and modulated waveforms.

        Args:
            savefig_filename (str): a file path. If provided the plot is save to a file.
        """

        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        num_samples = int(np.rint(self.duration / 1e9 * PulseShape.SAMPLING_RATE))
        time = self.start + np.arange(num_samples) / PulseShape.SAMPLING_RATE * 1e9
        fig = plt.figure(figsize=(14, 5), dpi=200)
        gs = gridspec.GridSpec(ncols=2, nrows=1, width_ratios=[2, 1])
        ax1 = plt.subplot(gs[0])
        ax1.plot(
            time,
            self.shape.envelope_waveform_i.data,
            label="envelope i",
            c="C0",
            linestyle="dashed",
        )
        ax1.plot(
            time,
            self.shape.envelope_waveform_q.data,
            label="envelope q",
            c="C1",
            linestyle="dashed",
        )
        ax1.plot(time, self.shape.modulated_waveform_i.data, label="modulated i", c="C0")
        ax1.plot(time, self.shape.modulated_waveform_q.data, label="modulated q", c="C1")
        ax1.plot(time, -self.shape.envelope_waveform_i.data, c="silver", linestyle="dashed")
        ax1.set_xlabel("Time [ns]")
        ax1.set_ylabel("Amplitude")

        ax1.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
        ax1.axis([self.start, self.finish, -1, 1])
        ax1.legend()

        ax2 = plt.subplot(gs[1])
        ax2.plot(
            self.shape.modulated_waveform_i.data,
            self.shape.modulated_waveform_q.data,
            label="modulated",
            c="C3",
        )
        ax2.plot(
            self.shape.envelope_waveform_i.data,
            self.shape.envelope_waveform_q.data,
            label="envelope",
            c="C2",
        )
        ax2.plot(
            self.shape.modulated_waveform_i.data[0],
            self.shape.modulated_waveform_q.data[0],
            marker="o",
            markersize=5,
            label="start",
            c="lightcoral",
        )
        ax2.plot(
            self.shape.modulated_waveform_i.data[-1],
            self.shape.modulated_waveform_q.data[-1],
            marker="o",
            markersize=5,
            label="finish",
            c="darkred",
        )

        ax2.plot(
            np.cos(time * 2 * np.pi / self.duration),
            np.sin(time * 2 * np.pi / self.duration),
            c="silver",
            linestyle="dashed",
        )

        ax2.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
        ax2.legend()
        # ax2.axis([ -1, 1, -1, 1])
        ax2.axis("equal")
        plt.suptitle(self.serial)
        if savefig_filename:
            plt.savefig(savefig_filename)
        else:
            plt.show()
        plt.close()


class ReadoutPulse(Pulse):
    """Describes a readout pulse.

    See :class:`qibolab.pulses.Pulse` for argument desciption.
    """

    def __init__(
        self,
        start,
        duration,
        amplitude,
        frequency,
        relative_phase,
        shape,
        channel,
        qubit=0,
    ):
        # def __init__(self, start:int | se_int, duration:int | se_int, amplitude:float, frequency:int, relative_phase:float, shape: PulseShape | str,
        #                    channel: int | str, qubit: int | str = 0):
        super().__init__(
            start,
            duration,
            amplitude,
            frequency,
            relative_phase,
            shape,
            channel,
            type=PulseType.READOUT,
            qubit=qubit,
        )

    @property
    def serial(self):
        return f"ReadoutPulse({self.start}, {self.duration}, {format(self.amplitude, '.6f').rstrip('0').rstrip('.')}, {format(self.frequency, '_')}, {format(self.relative_phase, '.6f').rstrip('0').rstrip('.')}, {self.shape}, {self.channel}, {self.qubit})"

    @property
    def global_phase(self):
        # readout pulses should have zero global phase so that we can
        # calculate probabilities in the i-q plane
        return 0

    def copy(self):  # -> Pulse|ReadoutPulse|DrivePulse|FluxPulse:
        """Returns a new Pulse object with the same attributes."""

        # return eval(self.serial)
        return ReadoutPulse(
            self.start,
            self.duration,
            self.amplitude,
            self.frequency,
            self.relative_phase,
            repr(self._shape),  # self._shape,
            self.channel,
            self.qubit,
        )


class DrivePulse(Pulse):
    """Describes a qubit drive pulse.

    See :class:`qibolab.pulses.Pulse` for argument desciption.
    """

    def __init__(
        self,
        start,
        duration,
        amplitude,
        frequency,
        relative_phase,
        shape,
        channel,
        qubit=0,
    ):
        # def __init__(self, start:int | se_int, duration:int | se_int, amplitude:float, frequency:int, relative_phase:float, shape: PulseShape | str,
        #                    channel: int | str, qubit: int | str = 0):
        super().__init__(
            start,
            duration,
            amplitude,
            frequency,
            relative_phase,
            shape,
            channel,
            type=PulseType.DRIVE,
            qubit=qubit,
        )

    @property
    def serial(self):
        return f"DrivePulse({self.start}, {self.duration}, {format(self.amplitude, '.6f').rstrip('0').rstrip('.')}, {format(self.frequency, '_')}, {format(self.relative_phase, '.6f').rstrip('0').rstrip('.')}, {self.shape}, {self.channel}, {self.qubit})"


class FluxPulse(Pulse):
    """Describes a qubit flux pulse.

    Flux pulses have frequency and relative_phase equal to 0. Their i and q components are equal.
    See :class:`qibolab.pulses.Pulse` for argument desciption.
    """

    def __init__(self, start, duration, amplitude, shape, channel, qubit=0):
        # def __init__(self, start:int | se_int, duration:int | se_int, amplitude:float, frequency:int, relative_phase:float, shape: PulseShape | str,
        #                    channel: int | str, qubit: int | str = 0):
        super().__init__(
            start,
            duration,
            amplitude,
            0,
            0,
            shape,
            channel,
            type=PulseType.FLUX,
            qubit=qubit,
        )

    @property
    def envelope_waveform_i(self) -> Waveform:
        return self._shape.envelope_waveform_i

    @property
    def envelope_waveform_q(self) -> Waveform:
        return self._shape.envelope_waveform_i

    @property
    def envelope_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        return (self._shape.envelope_waveform_i, self._shape.envelope_waveform_i)

    @property
    def modulated_waveform_i(self) -> Waveform:
        return self._shape.envelope_waveform_i

    @property
    def modulated_waveform_q(self) -> Waveform:
        return self._shape.envelope_waveform_i

    @property
    def modulated_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        return (self._shape.envelope_waveform_i, self._shape.envelope_waveform_i)

    @property
    def serial(self):
        return f"FluxPulse({self.start}, {self.duration}, {format(self.amplitude, '.6f').rstrip('0').rstrip('.')}, {self.shape}, {self.channel}, {self.qubit})"


class SplitPulse(Pulse):
    """A supporting class to represent sections or slices of a pulse."""

    # TODO: Since this class is only required by qblox drivers, move to qblox.py
    def __init__(self, pulse: Pulse, window_start: int = None, window_finish: int = None):
        super().__init__(
            pulse.start,
            pulse.duration,
            pulse.amplitude,
            pulse.frequency,
            pulse.relative_phase,
            eval(str(pulse.shape)),
            pulse.channel,
            type=pulse.type,
            qubit=pulse.qubit,
        )
        self._window_start: int = pulse.start
        self._window_finish: int = pulse.finish
        if not window_start:
            window_start = pulse.start
        if not window_finish:
            window_finish = pulse.finish
        self.window_start = window_start
        self.window_finish = window_finish

    @property
    def window_start(self):
        return self._window_start

    @window_start.setter
    def window_start(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f"window_start argument type should be int, got {type(value).__name__}")
        if value < self.start:
            raise ValueError("window_start should be >= pulse start ({self._start}), got {value}")
        self._window_start = value

    @property
    def window_finish(self):
        return self._window_finish

    @window_finish.setter
    def window_finish(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f"window_start argument type should be int, got {type(value).__name__}")
        if value > self.finish:
            raise ValueError("window_finish should be <= pulse finish ({self._finish}), got {value}")
        self._window_finish = value

    @property
    def window_duration(self):
        return self._window_finish - self._window_start

    @property
    def serial(self):
        return f"SplitPulse({self.window_start}, {self.window_duration}, {format(self.amplitude, '.6f').rstrip('0').rstrip('.')}, {format(self.frequency, '_')}, {format(self.relative_phase, '.6f').rstrip('0').rstrip('.')}, {self.shape}, {self.channel}, {self.qubit})"

    @property
    def envelope_waveform_i(self) -> Waveform:
        waveform = Waveform(
            self._shape.envelope_waveform_i.data[self._window_start - self.start : self._window_finish - self.start]
        )
        waveform.serial = (
            self._shape.envelope_waveform_i.serial
            + f"[{self._window_start - self.start} : {self._window_finish - self.start}]"
        )
        return waveform

    @property
    def envelope_waveform_q(self) -> Waveform:
        waveform = Waveform(
            self._shape.modulated_waveform_i.data[self._window_start - self.start : self._window_finish - self.start]
        )
        waveform.serial = (
            self._shape.modulated_waveform_i.serial
            + f"[{self._window_start - self.start} : {self._window_finish - self.start}]"
        )
        return waveform

    @property
    def envelope_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        return (self.envelope_waveform_i, self.envelope_waveform_q)

    @property
    def modulated_waveform_i(self) -> Waveform:
        waveform = Waveform(
            self._shape.envelope_waveform_q.data[self._window_start - self.start : self._window_finish - self.start]
        )
        waveform.serial = (
            self._shape.envelope_waveform_q.serial
            + f"[{self._window_start - self.start} : {self._window_finish - self.start}]"
        )
        return waveform

    @property
    def modulated_waveform_q(self) -> Waveform:
        waveform = Waveform(
            self._shape.modulated_waveform_q.data[self._window_start - self.start : self._window_finish - self.start]
        )
        waveform.serial = (
            self._shape.modulated_waveform_q.serial
            + f"[{self._window_start - self.start} : {self._window_finish - self.start}]"
        )
        return waveform

    @property
    def modulated_waveforms(self):  #  -> tuple[Waveform, Waveform]:
        return (self.modulated_waveform_i, self.modulated_waveform_q)

    def plot(self, savefig_filename=None):
        import matplotlib.pyplot as plt
        from matplotlib import gridspec

        time = (
            self.window_start
            + np.arange(int(np.rint(self.window_duration / 1e9 * PulseShape.SAMPLING_RATE)))
            / PulseShape.SAMPLING_RATE
            * 1e9
        )

        fig = plt.figure(figsize=(14, 5), dpi=200)
        gs = gridspec.GridSpec(ncols=2, nrows=1, width_ratios=[2, 1])
        ax1 = plt.subplot(gs[0])
        ax1.plot(
            time,
            self.shape.envelope_waveform_i.data[self._window_start - self.start : self._window_finish - self.start],
            label="envelope i",
            c="C0",
            linestyle="dashed",
        )
        ax1.plot(
            time,
            self.shape.envelope_waveform_q.data[self._window_start - self.start : self._window_finish - self.start],
            label="envelope q",
            c="C1",
            linestyle="dashed",
        )
        ax1.plot(
            time,
            self.shape.modulated_waveform_i.data[self._window_start - self.start : self._window_finish - self.start],
            label="modulated i",
            c="C0",
        )
        ax1.plot(
            time,
            self.shape.modulated_waveform_q.data[self._window_start - self.start : self._window_finish - self.start],
            label="modulated q",
            c="C1",
        )
        ax1.plot(
            time,
            -self.shape.envelope_waveform_i.data[self._window_start - self.start : self._window_finish - self.start],
            c="silver",
            linestyle="dashed",
        )
        ax1.set_xlabel("Time [ns]")
        ax1.set_ylabel("Amplitude")

        ax1.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
        ax1.axis([self.window_start, self._window_finish, -1, 1])
        ax1.legend()

        ax2 = plt.subplot(gs[1])
        ax2.plot(
            self.shape.modulated_waveform_i.data[self._window_start - self.start : self._window_finish - self.start],
            self.shape.modulated_waveform_q.data[self._window_start - self.start : self._window_finish - self.start],
            label="modulated",
            c="C3",
        )
        ax2.plot(
            self.shape.envelope_waveform_i.data[self._window_start - self.start : self._window_finish - self.start],
            self.shape.envelope_waveform_q.data[self._window_start - self.start : self._window_finish - self.start],
            label="envelope",
            c="C2",
        )
        ax2.plot(
            np.cos(time * 2 * np.pi / self.window_duration),
            np.sin(time * 2 * np.pi / self.window_duration),
            c="silver",
            linestyle="dashed",
        )

        ax2.grid(visible=True, which="both", axis="both", color="#888888", linestyle="-")
        ax2.legend()
        # ax2.axis([ -1, 1, -1, 1])
        ax2.axis("equal")
        if savefig_filename:
            plt.savefig(savefig_filename)
        else:
            plt.show()
        plt.close()


class PulseConstructor(Enum):
    """An enumeration to map each ``PulseType`` to the proper pulse constructor."""

    READOUT = ReadoutPulse
    DRIVE = DrivePulse
    FLUX = FluxPulse


class PulseSequence:
    """A collection of scheduled pulses.

    A quantum circuit can be translated into a set of scheduled pulses that implement the circuit gates.
    This class contains many supporting fuctions to facilitate the creation and manipulation of
    these collections of pulses.
    None of the methods of PulseSequence modify any of the properties of its pulses.
    """

    def __init__(self, *pulses):
        self.pulses = []  #: list[Pulse] = []
        """pulses (list): a list containing the pulses, ordered by their channel and start times."""
        self.add(*pulses)

    def __len__(self):
        return len(self.pulses)

    def __iter__(self):
        return iter(self.pulses)

    def __getitem__(self, index):
        return self.pulses[index]

    def __setitem__(self, index, value):
        self.pulses[index] = value

    def __delitem__(self, index):
        del self.pulses[index]

    def __contains__(self, pulse):
        return pulse in self.pulses

    def __repr__(self):
        return self.serial

    @property
    def serial(self):
        """Returns a string representation of the pulse sequence."""

        return "PulseSequence\n" + "\n".join(f"{pulse.serial}" for pulse in self.pulses)

    def __eq__(self, other):
        if not isinstance(other, PulseSequence):
            raise TypeError(f"Expected PulseSequence; got {type(other).__name__}")
        return self.serial == other.serial

    def __ne__(self, other):
        if not isinstance(other, PulseSequence):
            raise TypeError(f"Expected PulseSequence; got {type(other).__name__}")
        return self.serial != other.serial

    def __hash__(self):
        return hash(self.serial)

    def __add__(self, other):
        if isinstance(other, PulseSequence):
            return PulseSequence(*self.pulses, *other.pulses)
        if isinstance(other, Pulse):
            return PulseSequence(*self.pulses, other)
        raise TypeError(f"Expected PulseSequence or Pulse; got {type(other).__name__}")

    def __radd__(self, other):
        if isinstance(other, PulseSequence):
            return PulseSequence(*other.pulses, *self.pulses)
        if isinstance(other, Pulse):
            return PulseSequence(other, *self.pulses)
        raise TypeError(f"Expected PulseSequence or Pulse; got {type(other).__name__}")

    def __iadd__(self, other):
        if isinstance(other, PulseSequence):
            self.add(*other.pulses)
        elif isinstance(other, Pulse):
            self.add(other)
        else:
            raise TypeError(f"Expected PulseSequence or Pulse; got {type(other).__name__}")
        return self

    def __mul__(self, n):
        if not isinstance(n, int):
            raise TypeError(f"Expected int; got {type(n).__name__}")
        if n < 0:
            raise TypeError(f"argument n should be >=0, got {n}")
        return PulseSequence(*(self.pulses * n))

    def __rmul__(self, n):
        if not isinstance(n, int):
            raise TypeError(f"Expected int; got {type(n).__name__}")
        if n < 0:
            raise TypeError(f"argument n should be >=0, got {n}")
        return PulseSequence(*(self.pulses * n))

    def __imul__(self, n):
        if not isinstance(n, int):
            raise TypeError(f"Expected int; got {type(n).__name__}")
        if n < 1:
            raise TypeError(f"argument n should be >=1, got {n}")
        original_set = self.shallow_copy()
        for x in range(n - 1):
            self.add(*original_set.pulses)
        return self

    @property
    def count(self):
        """Returns the number of pulses in the sequence."""

        return len(self.pulses)

    def add(self, *items):
        """Adds pulses to the sequence and sorts them by channel and start time."""

        for item in items:
            if isinstance(item, Pulse):
                pulse = item
                self.pulses.append(pulse)
            elif isinstance(item, PulseSequence):
                ps = item
                for pulse in ps.pulses:
                    self.pulses.append(pulse)
        self.pulses.sort(key=lambda item: (item.start, item.channel))

    def index(self, pulse):
        """Returns the index of a pulse in the sequence."""

        return self.pulses.index(pulse)

    def pop(self, index=-1):
        """Returns the pulse with the index provided and removes it from the sequence."""

        return self.pulses.pop(index)

    def remove(self, pulse):
        """Removes a pulse from the sequence."""

        while pulse in self.pulses:
            self.pulses.remove(pulse)

    def clear(self):
        """Removes all pulses from the sequence."""

        self.pulses.clear()

    def shallow_copy(self):
        """Returns a shallow copy of the sequence.

        It returns a new PulseSequence object with references to the same Pulse objects.
        """

        return PulseSequence(*self.pulses)

    def copy(self):
        """Returns a deep copy of the sequence.

        It returns a new PulseSequence with replicates of each of the pulses contained in the original sequence.
        """

        return PulseSequence(*[pulse.copy() for pulse in self.pulses])

    @property
    def ro_pulses(self):
        """Returns a new PulseSequence containing only its readout pulses."""

        new_pc = PulseSequence()
        for pulse in self.pulses:
            if pulse.type == PulseType.READOUT:
                new_pc.add(pulse)
        return new_pc

    @property
    def qd_pulses(self):
        """Returns a new PulseSequence containing only its qubit drive pulses."""

        new_pc = PulseSequence()
        for pulse in self.pulses:
            if pulse.type == PulseType.DRIVE:
                new_pc.add(pulse)
        return new_pc

    @property
    def qf_pulses(self):
        """Returns a new PulseSequence containing only its qubit flux pulses."""

        new_pc = PulseSequence()
        for pulse in self.pulses:
            if pulse.type == PulseType.FLUX:
                new_pc.add(pulse)
        return new_pc

    def get_channel_pulses(self, *channels):
        """Returns a new PulseSequence containing only the pulses on a specific set of channels."""

        new_pc = PulseSequence()
        for pulse in self.pulses:
            if pulse.channel in channels:
                new_pc.add(pulse)
        return new_pc

    def get_qubit_pulses(self, *qubits):
        """Returns a new PulseSequence containing only the pulses on a specific set of qubits."""

        new_pc = PulseSequence()
        for pulse in self.pulses:
            if pulse.qubit in qubits:
                new_pc.add(pulse)
        return new_pc

    @property
    def is_empty(self):
        """Returns True if the sequence does not contain any pulses."""

        return len(self.pulses) == 0

    @property
    def finish(self) -> int:
        """Returns the time when the last pulse of the sequence finishes."""

        t: int = 0
        for pulse in self.pulses:
            if pulse.finish > t:
                t = pulse.finish
        return t

    @property
    def start(self) -> int:
        """Returns the start time of the first pulse of the sequence."""

        t = self.finish
        for pulse in self.pulses:
            if pulse.start < t:
                t = pulse.start
        return t

    @property
    def duration(self) -> int:
        """Returns duration of the sequence calculated as its finish - start times."""

        return self.finish - self.start

    @property
    def channels(self) -> list:
        """Returns list containing the channels used by the pulses in the sequence."""

        channels = []
        for pulse in self.pulses:
            if not pulse.channel in channels:
                channels.append(pulse.channel)
        channels.sort()
        return channels

    @property
    def qubits(self) -> list:
        """Returns list containing the qubits associated with the pulses in the sequence."""

        qubits = []
        for pulse in self.pulses:
            if not pulse.qubit in qubits:
                qubits.append(pulse.qubit)
        qubits.sort()
        return qubits

    def get_pulse_overlaps(self):  # -> dict((int,int): PulseSequence):
        """Returns a dictionary of slices of time (tuples with start and finish times) where pulses overlap."""

        times = []
        for pulse in self.pulses:
            if not pulse.start in times:
                times.append(pulse.start)
            if not pulse.finish in times:
                times.append(pulse.finish)
        times.sort()

        overlaps = {}
        for n in range(len(times) - 1):
            overlaps[(times[n], times[n + 1])] = PulseSequence()
            for pulse in self.pulses:
                if (pulse.start <= times[n]) & (pulse.finish >= times[n + 1]):
                    overlaps[(times[n], times[n + 1])] += pulse
        return overlaps

    def separate_overlapping_pulses(self):  # -> dict((int,int): PulseSequence):
        """Separates a sequence of overlapping pulses into a list of non-overlapping sequences."""

        # This routine separates the pulses of a sequence into non-overlapping sets
        # but it does not check if the frequencies of the pulses within a set have the same frequency

        separated_pulses = []
        for new_pulse in self.pulses:
            stored = False
            for ps in separated_pulses:
                overlaps = False
                for existing_pulse in ps:
                    if new_pulse.start < existing_pulse.finish and new_pulse.finish > existing_pulse.start:
                        overlaps = True
                        break
                if not overlaps:
                    ps.add(new_pulse)
                    stored = True
                    break
            if not stored:
                separated_pulses.append(PulseSequence(new_pulse))
        return separated_pulses

    # TODO: Implement separate_different_frequency_pulses()

    @property
    def pulses_overlap(self) -> bool:
        """Returns True if any of the pulses in the sequence overlap."""

        overlap = False
        for pc in self.get_pulse_overlaps().values():
            if pc.count > 1:
                overlap = True
        return overlap

    def plot(self, savefig_filename=None):
        """Plots the sequence of pulses.

        Args:
            savefig_filename (str): a file path. If provided the plot is save to a file.
        """

        if not self.is_empty:
            import matplotlib.pyplot as plt
            from matplotlib import gridspec

            fig = plt.figure(figsize=(14, 2 * self.count), dpi=200)
            gs = gridspec.GridSpec(ncols=1, nrows=self.count)
            vertical_lines = []
            for pulse in self.pulses:
                vertical_lines.append(pulse.start)
                vertical_lines.append(pulse.finish)

            n = -1
            for qubit in self.qubits:
                qubit_pulses = self.get_qubit_pulses(qubit)
                for channel in qubit_pulses.channels:
                    n += 1
                    channel_pulses = qubit_pulses.get_channel_pulses(channel)
                    ax = plt.subplot(gs[n])
                    ax.axis([0, self.finish, -1, 1])
                    for pulse in channel_pulses:
                        if isinstance(pulse, SplitPulse):
                            num_samples = int(pulse.window_duration / 1e9 * PulseShape.SAMPLING_RATE)
                            time = pulse.window_start + np.arange(num_samples) / PulseShape.SAMPLING_RATE * 1e9
                            ax.plot(
                                time,
                                pulse.shape.modulated_waveform_q.data[
                                    pulse.window_start - pulse.start : pulse.window_finish - pulse.start
                                ],
                                c="lightgrey",
                            )
                            ax.plot(
                                time,
                                pulse.shape.modulated_waveform_i.data[
                                    pulse.window_start - pulse.start : pulse.window_finish - pulse.start
                                ],
                                c=f"C{str(n)}",
                            )
                            ax.plot(
                                time,
                                pulse.shape.envelope_waveform_i.data[
                                    pulse.window_start - pulse.start : pulse.window_finish - pulse.start
                                ],
                                c=f"C{str(n)}",
                            )
                            ax.plot(
                                time,
                                -pulse.shape.envelope_waveform_i.data[
                                    pulse.window_start - pulse.start : pulse.window_finish - pulse.start
                                ],
                                c=f"C{str(n)}",
                            )
                        else:
                            num_samples = int(pulse.duration / 1e9 * PulseShape.SAMPLING_RATE)
                            time = pulse.start + np.arange(num_samples) / PulseShape.SAMPLING_RATE * 1e9
                            ax.plot(
                                time,
                                pulse.shape.modulated_waveform_q.data,
                                c="lightgrey",
                            )
                            ax.plot(
                                time,
                                pulse.shape.modulated_waveform_i.data,
                                c=f"C{str(n)}",
                            )
                            ax.plot(
                                time,
                                pulse.shape.envelope_waveform_i.data,
                                c=f"C{str(n)}",
                            )
                            ax.plot(
                                time,
                                -pulse.shape.envelope_waveform_i.data,
                                c=f"C{str(n)}",
                            )
                        # TODO: if they overlap use different shades
                        ax.axhline(0, c="dimgrey")
                        ax.set_ylabel(f"qubit {qubit} \n channel {channel}")
                        for vl in vertical_lines:
                            ax.axvline(vl, c="slategrey", linestyle="--")
                        ax.axis([0, self.finish, -1, 1])
                        ax.grid(
                            visible=True,
                            which="both",
                            axis="both",
                            color="#CCCCCC",
                            linestyle="-",
                        )
            if savefig_filename:
                plt.savefig(savefig_filename)
            else:
                plt.show()
            plt.close()



from dataclasses import dataclass
from typing import List

from qibo.config import log

from qibolab.transpilers.abstract import Transpiler
from qibolab.transpilers.fusion import Fusion, Rearrange
from qibolab.transpilers.gate_decompositions import NativeGates
from qibolab.transpilers.star_connectivity import StarConnectivity


@dataclass
class Pipeline(Transpiler):
    """Transpiler consisting of a list of smaller transpilers that are applied sequentially."""

    transpilers: List[Transpiler]
    verbose: bool = False

    def is_satisfied(self, circuit):
        for transpiler in self.transpilers:
            if not transpiler.is_satisfied(circuit):
                return False
        return True

    def __call__(self, circuit):
        if self.verbose:
            log.info("Transpiling circuit.")
        nqubits = circuit.nqubits
        total_map = list(range(nqubits))
        for transpiler in self.transpilers:
            circuit, qubit_map = transpiler(circuit)
            total_map = [qubit_map[total_map[i]] for i in range(nqubits)]
        return circuit, total_map

    @classmethod
    def default(cls, two_qubit_natives, middle_qubit=2, fuse_one_qubit=False, verbose=False):
        """Default transpiler used by :class:`qibolab.backends.QibolabBackend`."""
        transpilers = [Rearrange(max_qubits=2)]
        # add SWAPs to satisfy connectivity constraints
        transpilers.append(StarConnectivity(middle_qubit))
        # two-qubit gates to native
        transpilers.append(NativeGates(two_qubit_natives, translate_single_qubit=False, verbose=verbose))
        # Optional: fuse one-qubit gates to reduce circuit depth
        if fuse_one_qubit:
            transpilers.append(Fusion(max_qubits=1))
        # one-qubit gates to native
        transpilers.append(NativeGates(two_qubit_natives, translate_single_qubit=True, verbose=verbose))
        return cls(transpilers, verbose)



from collections import defaultdict
from dataclasses import dataclass, field

from qibo import gates
from qibo.config import raise_error

from qibolab.compilers.default import (
    cz_rule,
    gpi2_rule,
    identity_rule,
    measurement_rule,
    rz_rule,
    u3_rule,
    z_rule,
)
from qibolab.pulses import PulseSequence, ReadoutPulse


@dataclass
class Compiler:
    """Compiler that transforms a :class:`qibo.models.Circuit` to a :class:`qibolab.pulses.PulseSequence`.

    The transformation is done using a dictionary of rules which map each Qibo gate to a
    pulse sequence and some virtual Z-phases.

    A rule is a function that takes two argumens:
        - gate (:class:`qibo.gates.abstract.Gate`): Gate object to be compiled.
        - platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): Platform object to read
            native gate pulses from.

    and returns:
        - sequence (:class:`qibolab.pulses.PulseSequence`): Sequence of pulses that implement
            the given gate.
        - virtual_z_phases (dict): Dictionary mapping qubits to virtual Z-phases induced by the gate.

    See :class:`qibolab.compilers.default` for an example of a compiler implementation.
    """

    rules: dict = field(default_factory=dict)
    """Map from gates to compilation rules."""

    @classmethod
    def default(cls):
        return cls(
            {
                gates.I: identity_rule,
                gates.Z: z_rule,
                gates.RZ: rz_rule,
                gates.U3: u3_rule,
                gates.CZ: cz_rule,
                gates.GPI2: gpi2_rule,
                gates.M: measurement_rule,
            }
        )

    def __setitem__(self, key, rule):
        """Sets a new rule to the compiler.

        If a rule already exists for the gate, it will be overwritten.
        """
        self.rules[key] = rule

    def __getitem__(self, item):
        """Get an existing rule for a given gate."""
        try:
            return self.rules[item]
        except KeyError:
            raise_error(KeyError, f"Compiler rule not available for {item}.")

    def __delitem__(self, item):
        """Remove rule for the given gate."""
        try:
            del self.rules[item]
        except KeyError:
            raise_error(KeyError, f"Cannot remove {item} from compiler because it does not exist.")

    def register(self, gate_cls):
        """Decorator for registering a function as a rule in the compiler.

        Using this decorator is optional. Alternatively the user can set the rules directly
        via ``__setitem__``.

        Args:
            gate_cls: Qibo gate object that the rule will be assigned to.
        """

        def inner(func):
            self[gate_cls] = func
            return func

        return inner

    def _compile_gate(self, gate, platform, sequence, virtual_z_phases, moment_start, delays):
        """Adds a single gate to the pulse sequence."""
        rule = self[gate.__class__]
        # get local sequence and phases for the current gate
        gate_sequence, gate_phases = rule(gate, platform)

        # update global pulse sequence
        # determine the right start time based on the availability of the qubits involved
        all_qubits = {*gate_sequence.qubits, *gate.qubits}
        start = max(*[sequence.get_qubit_pulses(qubit).finish + delays[qubit] for qubit in all_qubits], moment_start)
        # shift start time and phase according to the global sequence
        for pulse in gate_sequence:
            pulse.start += start
            if not isinstance(pulse, ReadoutPulse):
                pulse.relative_phase += virtual_z_phases[pulse.qubit]
            sequence.add(pulse)

        return gate_sequence, gate_phases

    def compile(self, circuit, platform):
        """Transforms a circuit to pulse sequence.

        Args:
            circuit (qibo.models.Circuit): Qibo circuit that respects the platform's
                                           connectivity and native gates.
            platform (qibolab.platforms.abstract.AbstractPlatform): Platform used
                to load the native pulse representations.

        Returns:
            sequence (qibolab.pulses.PulseSequence): Pulse sequence that implements the circuit.
            measurement_map (dict): Map from each measurement gate to the sequence of  readout pulse implementing it.
        """
        sequence = PulseSequence()
        # FIXME: This will not work with qubits that have string names
        # TODO: Implement a mapping between circuit qubit ids and platform ``Qubit``s
        virtual_z_phases = defaultdict(int)

        measurement_map = {}
        # process circuit gates
        delays = defaultdict(int)
        for moment in circuit.queue.moments:
            moment_start = sequence.finish
            for gate in set(filter(lambda x: x is not None, moment)):
                if isinstance(gate, gates.Align):
                    for qubit in gate.qubits:
                        delays[qubit] += gate.delay
                    continue
                gate_sequence, gate_phases = self._compile_gate(
                    gate, platform, sequence, virtual_z_phases, moment_start, delays
                )
                for qubit in gate.qubits:
                    delays[qubit] = 0

                # update virtual Z phases
                for qubit, phase in gate_phases.items():
                    virtual_z_phases[qubit] += phase

                # register readout sequences to ``measurement_map`` so that we can
                # properly map acquisition results to measurement gates
                if isinstance(gate, gates.M):
                    measurement_map[gate] = gate_sequence

        return sequence, measurement_map



from qibolab.channels import Channel, ChannelMap
from qibolab.instruments.dummy import DummyInstrument
from qibolab.instruments.oscillator import LocalOscillator
from qibolab.platform import Platform
from qibolab.serialize import load_qubits, load_settings

NAME = "dummy"
RUNCARD = {
    "nqubits": 4,
    "qubits": [
        0,
        1,
        2,
        3,
    ],
    "settings": {"sampling_rate": 1000000000, "relaxation_time": 0, "nshots": 1024},
    "topology": [[0, 1], [1, 2], [0, 3]],
    "native_gates": {
        "single_qubit": {
            0: {
                "RX": {
                    "duration": 40,
                    "amplitude": 0.005,
                    "frequency": 4700000000,
                    "shape": "Gaussian(5)",
                    "type": "qd",
                    "start": 0,
                    "phase": 0,
                },
                "MZ": {
                    "duration": 1000,
                    "amplitude": 0.0025,
                    "frequency": 7226500000,
                    "shape": "Rectangular()",
                    "type": "ro",
                    "start": 0,
                    "phase": 0,
                },
            },
            1: {
                "RX": {
                    "duration": 40,
                    "amplitude": 0.0484,
                    "frequency": 4855663000,
                    "shape": "Drag(5, -0.02)",
                    "type": "qd",
                    "start": 0,
                    "phase": 0,
                },
                "MZ": {
                    "duration": 620,
                    "amplitude": 0.003575,
                    "frequency": 7453265000,
                    "shape": "Rectangular()",
                    "type": "ro",
                    "start": 0,
                    "phase": 0,
                },
            },
            2: {
                "RX": {
                    "duration": 40,
                    "amplitude": 0.005,
                    "frequency": 2700000000,
                    "shape": "Gaussian(5)",
                    "type": "qd",
                    "start": 0,
                    "phase": 0,
                },
                "MZ": {
                    "duration": 1000,
                    "amplitude": 0.0025,
                    "frequency": 5226500000,
                    "shape": "Rectangular()",
                    "type": "ro",
                    "start": 0,
                    "phase": 0,
                },
            },
            3: {
                "RX": {
                    "duration": 40,
                    "amplitude": 0.0484,
                    "frequency": 5855663000,
                    "shape": "Drag(5, -0.02)",
                    "type": "qd",
                    "start": 0,
                    "phase": 0,
                },
                "MZ": {
                    "duration": 620,
                    "amplitude": 0.003575,
                    "frequency": 8453265000,
                    "shape": "Rectangular()",
                    "type": "ro",
                    "start": 0,
                    "phase": 0,
                },
            },
        },
        "two_qubit": {
            "0-1": {
                "CZ": [
                    {
                        "duration": 30,
                        "amplitude": 0.055,
                        "shape": "Rectangular()",
                        "qubit": 1,
                        "relative_start": 0,
                        "type": "qf",
                    },
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 1},
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 0},
                ]
            },
            "1-2": {
                "CZ": [
                    {
                        "duration": 30,
                        "amplitude": 0.055,
                        "shape": "Rectangular()",
                        "qubit": 1,
                        "relative_start": 0,
                        "type": "qf",
                    },
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 2},
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 1},
                ]
            },
            "0-3": {
                "CZ": [
                    {
                        "duration": 34,
                        "amplitude": 0.055,
                        "shape": "SNZ(1)",
                        "qubit": 3,
                        "relative_start": 0,
                        "type": "qf",
                    },
                    {
                        "duration": 4,
                        "amplitude": 0.055,
                        "shape": "SNZ(1)",
                        "qubit": 3,
                        "relative_start": 14,
                        "type": "qf",
                    },
                    {
                        "duration": 34,
                        "amplitude": 0.1,
                        "shape": "Rectangular()",
                        "qubit": 3,
                        "relative_start": 0,
                        "type": "qf",
                    },
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 3},
                    {"type": "virtual_z", "phase": -1.5707963267948966, "qubit": 0},
                ]
            },
        },
    },
    "characterization": {
        "single_qubit": {
            0: {
                "readout_frequency": 7226500000,
                "drive_frequency": 4700000000,
                "T1": 0.0,
                "T2": 0.0,
                "sweetspot": 0.0,
                "threshold": 0.0,
                "iq_angle": 0.0,
            },
            1: {
                "readout_frequency": 7453265000,
                "drive_frequency": 4855663000,
                "T1": 0.0,
                "T2": 0.0,
                "sweetspot": -0.047,
                "threshold": 0.00028502261712637096,
                "iq_angle": 1.283105298787488,
            },
            2: {
                "readout_frequency": 5226500000,
                "drive_frequency": 2700000000,
                "T1": 0.0,
                "T2": 0.0,
                "sweetspot": 0.0,
                "threshold": 0.0,
                "iq_angle": 0.0,
            },
            3: {
                "readout_frequency": 8453265000,
                "drive_frequency": 5855663000,
                "T1": 0.0,
                "T2": 0.0,
                "sweetspot": -0.047,
                "threshold": 0.00028502261712637096,
                "iq_angle": 1.283105298787488,
            },
        }
    },
}


def create_dummy():
    """Create a dummy platform on two qubits using the dummy instrument."""
    # Create dummy controller
    instrument = DummyInstrument(NAME, 0)

    # Create local oscillator
    twpa_pump = LocalOscillator(name="twpa_pump", address=0)
    twpa_pump.frequency = 1e9
    twpa_pump.power = 10

    # Create channel objects
    nqubits = RUNCARD["nqubits"]
    channels = ChannelMap()
    channels |= Channel("readout", port=instrument["readout"])
    channels |= (Channel(f"drive-{i}", port=instrument[f"drive-{i}"]) for i in range(nqubits))
    channels |= (Channel(f"flux-{i}", port=instrument[f"flux-{i}"]) for i in range(nqubits))
    channels |= Channel("twpa", port=None)
    channels["readout"].attenuation = 0
    channels["twpa"].local_oscillator = twpa_pump

    qubits, pairs = load_qubits(RUNCARD)
    settings = load_settings(RUNCARD)

    # map channels to qubits
    for q, qubit in qubits.items():
        qubit.readout = channels["readout"]
        qubit.drive = channels[f"drive-{q}"]
        qubit.flux = channels[f"flux-{q}"]
        qubit.twpa = channels["twpa"]

    instruments = {instrument.name: instrument, twpa_pump.name: twpa_pump}
    instrument.sampling_rate = settings.sampling_rate * 1e-9
    return Platform(NAME, qubits, pairs, instruments, settings, resonator_type="2D")



import importlib.metadata as im
import importlib.util
import os
from pathlib import Path

from qibo.config import raise_error

from qibolab.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)
from qibolab.platform import Platform

__version__ = im.version(__package__)

PLATFORMS = "QIBOLAB_PLATFORMS"


def get_platforms_path():
    """Get path to repository containing the platforms.

    Path is specified using the environment variable QIBOLAB_PLATFORMS.
    """
    profiles = os.environ.get(PLATFORMS)
    if profiles is None or not os.path.exists(profiles):
        raise_error(RuntimeError, f"Profile directory {profiles} does not exist.")
    return Path(profiles)


def create_platform(name, runcard=None):
    """A platform for executing quantum algorithms.

    It consists of a quantum processor QPU and a set of controlling instruments.

    Args:
        name (str): name of the platform. Options are 'tiiq', 'qili' and 'icarusq'.
    Returns:
        The plaform class.
    """
    if name == "dummy":
        from qibolab.dummy import create_dummy

        return create_dummy()

    platform = get_platforms_path() / f"{name}.py"
    if not platform.exists():
        raise_error(ValueError, f"Platform {name} does not exist.")

    spec = importlib.util.spec_from_file_location("platform", platform)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    if runcard is None:
        return module.create()
    return module.create(runcard)



"""Implementation of the default compiler.

Uses I, Z, RZ, U3, CZ, and M as the set of native gates.
"""
import math

from qibolab.pulses import PulseSequence


def identity_rule(gate, platform):
    """Identity gate skipped."""
    return PulseSequence(), {}


def z_rule(gate, platform):
    """Z gate applied virtually."""
    qubit = gate.target_qubits[0]
    return PulseSequence(), {qubit: math.pi}


def rz_rule(gate, platform):
    """RZ gate applied virtually."""
    qubit = gate.target_qubits[0]
    return PulseSequence(), {qubit: gate.parameters[0]}


def gpi2_rule(gate, platform):
    """Rule for GPI2."""
    qubit = gate.target_qubits[0]
    theta = gate.parameters[0]

    sequence = PulseSequence()
    pulse = platform.create_RX90_pulse(qubit, start=0, relative_phase=theta)
    sequence.add(pulse)
    return sequence, {}


def u3_rule(gate, platform):
    """U3 applied as RZ-RX90-RZ-RX90-RZ."""
    qubit = gate.target_qubits[0]
    # Transform gate to U3 and add pi/2-pulses
    theta, phi, lam = gate.parameters
    # apply RZ(lam)
    virtual_z_phases = {qubit: lam}
    sequence = PulseSequence()
    # Fetch pi/2 pulse from calibration
    RX90_pulse_1 = platform.create_RX90_pulse(qubit, start=0, relative_phase=virtual_z_phases[qubit])
    # apply RX(pi/2)
    sequence.add(RX90_pulse_1)
    # apply RZ(theta)
    virtual_z_phases[qubit] += theta
    # Fetch pi/2 pulse from calibration
    RX90_pulse_2 = platform.create_RX90_pulse(
        qubit, start=RX90_pulse_1.finish, relative_phase=virtual_z_phases[qubit] - math.pi
    )
    # apply RX(-pi/2)
    sequence.add(RX90_pulse_2)
    # apply RZ(phi)
    virtual_z_phases[qubit] += phi

    return sequence, virtual_z_phases


def cz_rule(gate, platform):
    """CZ applied as defined in the platform runcard.

    Applying the CZ gate may involve sending pulses on qubits
    that the gate is not directly acting on.
    """
    return platform.create_CZ_pulse_sequence(gate.qubits)


def measurement_rule(gate, platform):
    """Measurement gate applied using the platform readout pulse."""
    sequence = PulseSequence()
    for qubit in gate.target_qubits:
        MZ_pulse = platform.create_MZ_pulse(qubit, start=0)
        sequence.add(MZ_pulse)
    return sequence, {}



from dataclasses import dataclass

import numpy as np
from qibo.config import log, raise_error

from qibolab.instruments.port import Port

FREQUENCY_LIMIT = 500e6


@dataclass
class QbloxOutputPort_Settings:
    channel: str = None
    qubit: str = None
    attenuation: int = 60
    offset: float = 0.0
    hardware_mod_en: bool = True
    gain: float = 1.0
    nco_freq: int = 0
    nco_phase_offs: float = 0


@dataclass
class ClusterRF_OutputPort_Settings(QbloxOutputPort_Settings):
    lo_enabled: bool = True
    lo_frequency: int = 2_000_000_000


@dataclass
class ClusterBB_OutputPort_Settings(QbloxOutputPort_Settings):
    pass


@dataclass
class QbloxInputPort_Settings:
    channel: str = None
    acquisition_hold_off: int = 0
    acquisition_duration: int = 1000
    hardware_demod_en: bool = True


class QbloxOutputPort(Port):
    """qibolab.instruments.port.Port interface implementation for Qblox instruments"""

    def __init__(self, module, sequencer_number: int, number: int):
        self.module = module
        self.sequencer_number: int = sequencer_number
        self.port_number: int = number - 1
        self.channel = None  # To be discontinued
        self.qubit = None  # To be discontinued

        self._settings = QbloxOutputPort_Settings()

    @property
    def attenuation(self) -> str:
        """Attenuation that is applied to this port."""
        if self.module.device:
            self._settings.attenuation = self.module.device.get(f"out{self.port_number}_att")
        return self._settings.attenuation

    @attenuation.setter
    def attenuation(self, value):
        if isinstance(value, (float, np.floating)):
            value = int(value)
        if isinstance(value, (int, np.integer)):
            if value > 60:
                log.warning(f"Qblox attenuation needs to be between 0 and 60 dB. Adjusting {value} to 60dB")
                value = 60

            elif value < 0:
                log.warning(f"Qblox attenuation needs to be between 0 and 60 dB. Adjusting {value} to 0")
                value = 0

            if (value % 2) != 0:
                log.warning(
                    f"Qblox attenuation needs to be a multiple of 2 dB. Adjusting {value} to {round(value/2) * 2}"
                )
                value = round(value / 2) * 2
        else:
            raise_error(ValueError, f"Invalid attenuation {value}")

        self._settings.attenuation = value
        if self.module.device:
            self.module._set_device_parameter(self.module.device, f"out{self.port_number}_att", value=value)

    @property
    def offset(self):
        """DC offset that is applied to this port."""
        if self.module.device:
            self._settings.offset = self.module.device.get(f"out{self.port_number}_offset")
        return self._settings.offset

    @offset.setter
    def offset(self, value):
        if isinstance(value, (int, np.integer)):
            value = float(value)
        if isinstance(value, (float, np.floating)):
            if value > 2.5:
                log.warning(f"Qblox offset needs to be between -2.5 and 2.5 V. Adjusting {value} to 2.5 V")
                value = 2.5

            elif value < -2.5:
                log.warning(f"Qblox offset needs to be between -2.5 and 2.5 V. Adjusting {value} to -2.5 V")
                value = -2.5
        else:
            raise_error(ValueError, f"Invalid offset {value}")

        self._settings.offset = value
        if self.module.device:
            self.module._set_device_parameter(self.module.device, f"out{self.port_number}_offset", value=value)

    # Additional attributes needed by the driver
    @property
    def hardware_mod_en(self):
        """Flag to enable hardware modulation."""
        if self.module.device:
            self._settings.hardware_mod_en = self.module.device.sequencers[self.sequencer_number].get(f"mod_en_awg")
        return self._settings.hardware_mod_en

    @hardware_mod_en.setter
    def hardware_mod_en(self, value):
        if not isinstance(value, bool):
            raise_error(ValueError, f"Invalid hardware_mod_en {value}")

        self._settings.hardware_mod_en = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.sequencer_number], "mod_en_awg", value=value
            )

    @property
    def nco_freq(self):
        """nco_freq that is applied to this port."""

        if self.module.device:
            self._settings.nco_freq = self.module.device.sequencers[self.sequencer_number].get(f"nco_freq")

        return self._settings.nco_freq

    @nco_freq.setter
    def nco_freq(self, value):
        if isinstance(value, (float, np.floating)):
            value = int(value)
        if isinstance(value, (int, np.integer)):
            if value > FREQUENCY_LIMIT:
                log.warning(
                    f"Qblox nco_freq needs to be between -{FREQUENCY_LIMIT} and {FREQUENCY_LIMIT} MHz. Adjusting {value} to {FREQUENCY_LIMIT} MHz"
                )
                value = int(FREQUENCY_LIMIT)

            elif value < -FREQUENCY_LIMIT:
                log.warning(
                    f"Qblox nco_freq needs to be between -{FREQUENCY_LIMIT} and {FREQUENCY_LIMIT} MHz. Adjusting {value} to -{FREQUENCY_LIMIT} MHz"
                )
                value = int(-FREQUENCY_LIMIT)
        else:
            raise_error(ValueError, f"Invalid nco_freq {value}")

        self._settings.nco_freq = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.sequencer_number], "nco_freq", value=value
            )

    @property
    def nco_phase_offs(self):
        """nco_phase_offs that is applied to this port."""

        if self.module.device:
            self._settings.nco_phase_offs = self.module.device.sequencers[self.sequencer_number].get(f"nco_phase_offs")
        return self._settings.nco_phase_offs

    @nco_phase_offs.setter
    def nco_phase_offs(self, value):
        if isinstance(value, (int, np.integer)):
            value = float(value)
        if isinstance(value, (float, np.floating)):
            value = value % 360
        else:
            raise_error(ValueError, f"Invalid nco_phase_offs {value}")

        self._settings.nco_phase_offs = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.sequencer_number], "nco_phase_offs", value=value
            )


class ClusterRF_OutputPort(QbloxOutputPort):
    def __init__(self, module, sequencer_number: int, number: int):
        super().__init__(module, sequencer_number, number)
        self._settings = ClusterRF_OutputPort_Settings()

    @property
    def lo_enabled(self):
        """Flag to enable local oscillator."""

        if self.module.device:
            if self.module.device.is_qrm_type:
                self._settings.lo_enabled = self.module.device.get(f"out{self.port_number}_in{self.port_number}_lo_en")
            elif self.module.device.is_qcm_type:
                self._settings.lo_enabled = self.module.device.get(f"out{self.port_number}_lo_en")
        return self._settings.lo_enabled

    @lo_enabled.setter
    def lo_enabled(self, value):
        if not isinstance(value, bool):
            raise_error(ValueError, f"Invalid lo_enabled {value}")

        self._settings.lo_enabled = value
        if self.module.device:
            if self.module.device.is_qrm_type:
                self.module._set_device_parameter(
                    self.module.device, f"out{self.port_number}_in{self.port_number}_lo_en", value=value
                )
            elif self.module.device.is_qcm_type:
                self.module._set_device_parameter(self.module.device, f"out{self.port_number}_lo_en", value=value)

    @property
    def lo_frequency(self):
        """Local oscillator frequency for the given port."""
        if self.module.device:
            if self.module.device.is_qrm_type:
                self._settings.lo_frequency = self.module.device.get(
                    f"out{self.port_number}_in{self.port_number}_lo_freq"
                )
            elif self.module.device.is_qcm_type:
                self._settings.lo_frequency = self.module.device.get(f"out{self.port_number}_lo_freq")
        return self._settings.lo_frequency

    @lo_frequency.setter
    def lo_frequency(self, value):
        if isinstance(value, (float, np.floating)):
            value = int(value)
        if isinstance(value, (int, np.integer)):
            if value > 18e9:
                log.warning(f"Qblox lo_frequency needs to be between 2e9 and 18e9 Hz. Adjusting {value} to 18e9 Hz")
                value = int(18e9)

            elif value < 2e9:
                log.warning(f"Qblox lo_frequency needs to be between 2e9 and 18e9 Hz. Adjusting {value} to 2e9 Hz")
                value = int(2e9)
        else:
            raise_error(ValueError, f"Invalid lo-frequency {value}")

        self._settings.lo_frequency = value
        if self.module.device:
            if self.module.device.is_qrm_type:
                self.module._set_device_parameter(
                    self.module.device, f"out{self.port_number}_in{self.port_number}_lo_freq", value=value
                )
            elif self.module.device.is_qcm_type:
                self.module._set_device_parameter(self.module.device, f"out{self.port_number}_lo_freq", value=value)

    # Note: for qublos, gain is equivalent to amplitude, since it does not bring any advantages
    # we plan to remove it soon.
    @property
    def gain(self):
        """Gain that is applied to this port."""

        if self.module.device:
            self._settings.gain = self.module.device.sequencers[self.sequencer_number].get(f"gain_awg_path0")
        return self._settings.gain

    @gain.setter
    def gain(self, value):
        if isinstance(value, (int, np.integer)):
            value = float(value)
        if isinstance(value, (float, np.floating)):
            if value > 1.0:
                log.warning(f"Qblox offset needs to be between -1 and 1. Adjusting {value} to 1")
                value = 1.0

            elif value < -1.0:
                log.warning(f"Qblox offset needs to be between -1 and 1. Adjusting {value} to -1")
                value = -1.0
        else:
            raise_error(ValueError, f"Invalid offset {value}")

        self._settings.gain = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.sequencer_number], "gain_awg_path0", "gain_awg_path1", value=value
            )


class ClusterBB_OutputPort(QbloxOutputPort):
    # Note: for qblox, gain is equivalent to amplitude, since it does not bring any advantages
    # we plan to remove it soon.

    def __init__(self, module, sequencer_number: int, number: int):
        super().__init__(module, sequencer_number, number)
        self._settings = ClusterBB_OutputPort_Settings()

    @property
    def gain(self):
        """Gain that is applied to this port."""

        if self.module.device:
            self._settings.gain = self.module.device.sequencers[self.sequencer_number].get(f"gain_awg_path0")
        return self._settings.gain

    @gain.setter
    def gain(self, value):
        if isinstance(value, (int, np.integer)):
            value = float(value)
        if isinstance(value, (float, np.floating)):
            if value > 1.0:
                log.warning(f"Qblox offset needs to be between -1 and 1. Adjusting {value} to 1")
                value = 1.0

            elif value < -1.0:
                log.warning(f"Qblox offset needs to be between -1 and 1. Adjusting {value} to -1")
                value = -1.0
        else:
            raise_error(ValueError, f"Invalid offset {value}")

        self._settings.gain = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.sequencer_number], "gain_awg_path0", value=value
            )


class QbloxInputPort:
    def __init__(self, module, output_sequencer_number: int, input_sequencer_number: int, number: int):
        self.module = module
        self.output_sequencer_number: int = output_sequencer_number
        self.input_sequencer_number: int = input_sequencer_number
        self.port_number: int = number - 1
        self.channel = None  # To be discontinued
        self.qubit = None  # To be discontinued

        self.acquisition_hold_off = 4  # To be discontinued

        self._settings = QbloxInputPort_Settings()

    @property
    def hardware_demod_en(self):
        """Flag to enable hardware demodulation."""

        if self.module.device:
            self._settings.hardware_demod_en = self.module.device.sequencers[self.input_sequencer_number].get(
                f"demod_en_acq"
            )
        return self._settings.hardware_demod_en

    @hardware_demod_en.setter
    def hardware_demod_en(self, value):
        if not isinstance(value, bool):
            raise_error(ValueError, f"Invalid hardware_demod_en {value}")

        self._settings.hardware_demod_en = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.input_sequencer_number], "demod_en_acq", value=value
            )

    @property
    def acquisition_duration(self):
        """Duration of the pulse acquisition, in ns. It must be > 0 and multiple of 4."""

        if self.module.device:
            self._settings.acquisition_duration = self.module.device.sequencers[self.output_sequencer_number].get(
                f"integration_length_acq"
            )
        return self._settings.acquisition_duration

    @acquisition_duration.setter
    def acquisition_duration(self, value):
        if isinstance(value, (float, np.floating)):
            value = int(value)
        if isinstance(value, (int, np.integer)):
            if value < 4:
                log.warning(f"Qblox hardware_demod_en needs to be > 4ns. Adjusting {value} to 4 ns")
                value = 4
            if (value % 4) != 0:
                log.warning(
                    f"Qblox hardware_demod_en needs to be a multiple of 4 ns. Adjusting {value} to {round(value/4) * 4}"
                )
                value = round(value / 4) * 4

        else:
            raise_error(ValueError, f"Invalid acquisition_duration {value}")

        self._settings.acquisition_duration = value
        if self.module.device:
            self.module._set_device_parameter(
                self.module.device.sequencers[self.output_sequencer_number], "integration_length_acq", value=value
            )



import operator
from dataclasses import dataclass
from enum import Enum, auto
from functools import partial
from typing import Optional

import numpy.typing as npt


class Parameter(Enum):
    """Sweeping parameters."""

    frequency = auto()
    amplitude = auto()
    duration = auto()
    relative_phase = auto()
    start = auto()

    attenuation = auto()
    gain = auto()
    bias = auto()
    lo_frequency = auto()


FREQUENCY = Parameter.frequency
AMPLITUDE = Parameter.amplitude
DURATION = Parameter.duration
RELATIVE_PHASE = Parameter.relative_phase
START = Parameter.start
ATTENUATION = Parameter.attenuation
GAIN = Parameter.gain
BIAS = Parameter.bias


class SweeperType(Enum):
    """Type of the Sweeper"""

    ABSOLUTE = partial(lambda x, y=None: x)
    FACTOR = operator.mul
    OFFSET = operator.add


QubitParameter = {Parameter.bias, Parameter.attenuation, Parameter.gain}


@dataclass
class Sweeper:
    """Data structure for Sweeper object.

    This object is passed as an argument to the method :func:`qibolab.platforms.abstract.Platform.sweep`
    which enables the user to sweep a specific parameter for one or more pulses. For information on how to
    perform sweeps see :func:`qibolab.platforms.abstract.Platform.sweep`.

    Example:
        .. testcode::

            import numpy as np
            from qibolab.dummy import create_dummy
            from qibolab.sweeper import Sweeper, Parameter
            from qibolab.pulses import PulseSequence
            from qibolab import ExecutionParameters


            platform = create_dummy()
            sequence = PulseSequence()
            parameter = Parameter.frequency
            pulse = platform.create_qubit_readout_pulse(qubit=0, start=0)
            sequence.add(pulse)
            parameter_range = np.random.randint(10, size=10)
            sweeper = Sweeper(parameter, parameter_range, [pulse])
            platform.sweep(sequence, ExecutionParameters(), sweeper)

    Args:
        parameter (`qibolab.sweeper.Parameter`): parameter to be swept, possible choices are frequency, attenuation, amplitude, current and gain.
        _values (np.ndarray): sweep range. If the parameter of the sweep is a pulse parameter, if the sweeper type is not ABSOLUTE, the base value
            will be taken from the runcard pulse parameters. If the sweep parameter is Bias, the base value will be the sweetspot of the qubits.
        pulses (list) : list of `qibolab.pulses.Pulse` to be swept (optional).
        qubits (list): list of `qibolab.platforms.abstract.Qubit` to be swept (optional).
        type (SweeperType): can be ABSOLUTE (the sweeper range is swept directly),
            FACTOR (sweeper values are multiplied by base value), OFFSET (sweeper values are added
            to base value)
    """

    parameter: Parameter
    values: npt.NDArray
    pulses: Optional[list] = None
    qubits: Optional[list] = None
    type: Optional[SweeperType] = SweeperType.ABSOLUTE

    def __post_init__(self):
        if self.pulses is not None and self.qubits is not None:
            raise ValueError("Cannot use a sweeper on both pulses and qubits.")
        if self.pulses is not None and self.parameter in QubitParameter:
            raise ValueError(f"Cannot sweep {self.parameter} without specifying qubits.")
        if self.qubits is not None and self.parameter not in QubitParameter:
            raise ValueError(f"Cannot sweep {self.parameter} without specifying pulses.")
        if self.pulses is None and self.qubits is None:
            raise ValueError("Cannot use a sweeper without specifying pulses or qubits.")

    def get_values(self, base_value):
        """Convert sweeper values depending on the sweeper type"""
        return self.type.value(self.values, base_value)



import math
from dataclasses import dataclass, field
from typing import Dict, Optional, Tuple, Union

import numpy as np
from qibo.config import raise_error

from qibolab.instruments.port import Port
from qibolab.pulses import PulseType, Rectangular

PortId = Tuple[str, int]
"""Type for port definition, for example: ("con1", 2)."""
IQPortId = Union[Tuple[PortId], Tuple[PortId, PortId]]
"""Type for collections of IQ ports."""


@dataclass
class QMPort(Port):
    name: IQPortId
    offset: float = 0.0
    gain: int = 0
    filters: Optional[Dict[str, float]] = None


@dataclass
class QMConfig:
    """Configuration for communicating with the ``QuantumMachinesManager``."""

    version: int = 1
    controllers: dict = field(default_factory=dict)
    elements: dict = field(default_factory=dict)
    pulses: dict = field(default_factory=dict)
    waveforms: dict = field(default_factory=dict)
    digital_waveforms: dict = field(default_factory=lambda: {"ON": {"samples": [(1, 0)]}})
    integration_weights: dict = field(default_factory=dict)
    mixers: dict = field(default_factory=dict)

    def register_analog_output_controllers(self, port: QMPort):
        """Register controllers in the ``config``.

        Args:
            ports (QMPort): Port we are registering.
                Contains information about the controller and port number and
                some parameters (offset, gain, filter, etc.).
        """
        for con, port_number in port.name:
            if con not in self.controllers:
                self.controllers[con] = {"analog_outputs": {}}
            self.controllers[con]["analog_outputs"][port_number] = {"offset": port.offset}
            if port.filters is not None:
                self.controllers[con]["analog_outputs"][port_number]["filter"] = port.filters

    @staticmethod
    def iq_imbalance(g, phi):
        """Creates the correction matrix for the mixer imbalance caused by the gain and phase imbalances

        More information here:
        https://docs.qualang.io/libs/examples/mixer-calibration/#non-ideal-mixer

        Args:
            g (float): relative gain imbalance between the I & Q ports (unit-less).
                Set to 0 for no gain imbalance.
            phi (float): relative phase imbalance between the I & Q ports (radians).
                Set to 0 for no phase imbalance.
        """
        c = np.cos(phi)
        s = np.sin(phi)
        N = 1 / ((1 - g**2) * (2 * c**2 - 1))
        return [float(N * x) for x in [(1 - g) * c, (1 + g) * s, (1 - g) * s, (1 + g) * c]]

    def register_drive_element(self, qubit, intermediate_frequency=0):
        """Register qubit drive elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"drive{qubit.name}" not in self.elements:
            # register drive controllers
            self.register_analog_output_controllers(qubit.drive.port)
            # register element
            lo_frequency = math.floor(qubit.drive.local_oscillator.frequency)
            self.elements[f"drive{qubit.name}"] = {
                "mixInputs": {
                    "I": qubit.drive.port.name[0],
                    "Q": qubit.drive.port.name[1],
                    "lo_frequency": lo_frequency,
                    "mixer": f"mixer_drive{qubit.name}",
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
            }
            drive_g = qubit.mixer_drive_g
            drive_phi = qubit.mixer_drive_phi
            self.mixers[f"mixer_drive{qubit.name}"] = [
                {
                    "intermediate_frequency": intermediate_frequency,
                    "lo_frequency": lo_frequency,
                    "correction": self.iq_imbalance(drive_g, drive_phi),
                }
            ]
        else:
            self.elements[f"drive{qubit.name}"]["intermediate_frequency"] = intermediate_frequency
            self.mixers[f"mixer_drive{qubit.name}"][0]["intermediate_frequency"] = intermediate_frequency

    def register_readout_element(self, qubit, intermediate_frequency=0, time_of_flight=0, smearing=0):
        """Register resonator elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"readout{qubit.name}" not in self.elements:
            # register readout controllers
            self.register_analog_output_controllers(qubit.readout.port)
            # register feedback controllers
            controllers = self.controllers
            for con, port_number in qubit.feedback.port.name:
                if con not in controllers:
                    controllers[con] = {
                        "analog_outputs": {},
                        "digital_outputs": {
                            1: {},
                        },
                        "analog_inputs": {},
                    }
                if "digital_outputs" not in controllers[con]:
                    controllers[con]["digital_outputs"] = {
                        1: {},
                    }
                if "analog_inputs" not in controllers[con]:
                    controllers[con]["analog_inputs"] = {}
                controllers[con]["analog_inputs"][port_number] = {"offset": 0.0, "gain_db": qubit.feedback.port.gain}
            # register element
            lo_frequency = math.floor(qubit.readout.local_oscillator.frequency)
            self.elements[f"readout{qubit.name}"] = {
                "mixInputs": {
                    "I": qubit.readout.port.name[0],
                    "Q": qubit.readout.port.name[1],
                    "lo_frequency": lo_frequency,
                    "mixer": f"mixer_readout{qubit.name}",
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
                "outputs": {
                    "out1": qubit.feedback.port.name[0],
                    "out2": qubit.feedback.port.name[1],
                },
                "time_of_flight": time_of_flight,
                "smearing": smearing,
            }
            readout_g = qubit.mixer_readout_g
            readout_phi = qubit.mixer_readout_phi
            self.mixers[f"mixer_readout{qubit.name}"] = [
                {
                    "intermediate_frequency": intermediate_frequency,
                    "lo_frequency": lo_frequency,
                    "correction": self.iq_imbalance(readout_g, readout_phi),
                }
            ]
        else:
            self.elements[f"readout{qubit.name}"]["intermediate_frequency"] = intermediate_frequency
            self.mixers[f"mixer_readout{qubit.name}"][0]["intermediate_frequency"] = intermediate_frequency

    def register_flux_element(self, qubit, intermediate_frequency=0):
        """Register qubit flux elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"flux{qubit.name}" not in self.elements:
            # register controller
            self.register_analog_output_controllers(qubit.flux.port)
            # register element
            self.elements[f"flux{qubit.name}"] = {
                "singleInput": {
                    "port": qubit.flux.port.name[0],
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
            }
        else:
            self.elements[f"flux{qubit.name}"]["intermediate_frequency"] = intermediate_frequency

    def register_element(self, qubit, pulse, time_of_flight=0, smearing=0):
        if pulse.type is PulseType.DRIVE:
            # register drive element
            if_frequency = pulse.frequency - math.floor(qubit.drive.local_oscillator.frequency)
            self.register_drive_element(qubit, if_frequency)
            # register flux element (if available)
            if qubit.flux:
                self.register_flux_element(qubit)
        elif pulse.type is PulseType.READOUT:
            # register readout element (if it does not already exist)
            if_frequency = pulse.frequency - math.floor(qubit.readout.local_oscillator.frequency)
            self.register_readout_element(qubit, if_frequency, time_of_flight, smearing)
            # register flux element (if available)
            if qubit.flux:
                self.register_flux_element(qubit)
        else:
            # register flux element
            self.register_flux_element(qubit, pulse.frequency)

    def register_pulse(self, qubit, pulse):
        """Registers pulse, waveforms and integration weights in QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit that the pulse acts on.
            pulse (:class:`qibolab.pulses.Pulse`): Pulse object to register.

        Returns:
            element (str): Name of the element this pulse will be played on.
                Elements are a part of the QM config and are generated during
                instantiation of the Qubit objects. They are named as
                "drive0", "drive1", "flux0", "readout0", ...
        """
        if pulse.serial not in self.pulses:
            if pulse.type is PulseType.DRIVE:
                serial_i = self.register_waveform(pulse, "i")
                serial_q = self.register_waveform(pulse, "q")
                self.pulses[pulse.serial] = {
                    "operation": "control",
                    "length": pulse.duration,
                    "waveforms": {"I": serial_i, "Q": serial_q},
                }
                # register drive pulse in elements
                self.elements[f"drive{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            elif pulse.type is PulseType.FLUX:
                serial = self.register_waveform(pulse)
                self.pulses[pulse.serial] = {
                    "operation": "control",
                    "length": pulse.duration,
                    "waveforms": {
                        "single": serial,
                    },
                }
                # register flux pulse in elements
                self.elements[f"flux{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            elif pulse.type is PulseType.READOUT:
                serial_i = self.register_waveform(pulse, "i")
                serial_q = self.register_waveform(pulse, "q")
                self.register_integration_weights(qubit, pulse.duration)
                self.pulses[pulse.serial] = {
                    "operation": "measurement",
                    "length": pulse.duration,
                    "waveforms": {
                        "I": serial_i,
                        "Q": serial_q,
                    },
                    "integration_weights": {
                        "cos": f"cosine_weights{qubit.name}",
                        "sin": f"sine_weights{qubit.name}",
                        "minus_sin": f"minus_sine_weights{qubit.name}",
                    },
                    "digital_marker": "ON",
                }
                # register readout pulse in elements
                self.elements[f"readout{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            else:
                raise_error(TypeError, f"Unknown pulse type {pulse.type.name}.")

    def register_waveform(self, pulse, mode="i"):
        """Registers waveforms in QM config.

        QM supports two kinds of waveforms, examples:
            "zero_wf": {"type": "constant", "sample": 0.0}
            "x90_wf": {"type": "arbitrary", "samples": x90_wf.tolist()}

        Args:
            pulse (:class:`qibolab.pulses.Pulse`): Pulse object to read the waveform from.
            mode (str): "i" or "q" specifying which channel the waveform will be played.

        Returns:
            serial (str): String with a serialization of the waveform.
                Used as key to identify the waveform in the config.
        """
        # Maybe need to force zero q waveforms
        # if pulse.type.name == "READOUT" and mode == "q":
        #    serial = "zero_wf"
        #    if serial not in self.waveforms:
        #        self.waveforms[serial] = {"type": "constant", "sample": 0.0}
        if isinstance(pulse.shape, Rectangular):
            serial = f"constant_wf{pulse.amplitude}"
            if serial not in self.waveforms:
                self.waveforms[serial] = {"type": "constant", "sample": pulse.amplitude}
        else:
            waveform = getattr(pulse, f"envelope_waveform_{mode}")
            serial = waveform.serial
            if serial not in self.waveforms:
                self.waveforms[serial] = {"type": "arbitrary", "samples": waveform.data.tolist()}
        return serial

    def register_integration_weights(self, qubit, readout_len):
        """Registers integration weights in QM config.

        Args:
            qubit (:class:`qibolab.platforms.quantum_machines.Qubit`): Qubit
                object that the integration weights will be used for.
            readout_len (int): Duration of the readout pulse in ns.
        """
        angle = 0
        self.integration_weights.update(
            {
                f"cosine_weights{qubit.name}": {
                    "cosine": [(np.cos(angle), readout_len)],
                    "sine": [(-np.sin(angle), readout_len)],
                },
                f"sine_weights{qubit.name}": {
                    "cosine": [(np.sin(angle), readout_len)],
                    "sine": [(np.cos(angle), readout_len)],
                },
                f"minus_sine_weights{qubit.name}": {
                    "cosine": [(-np.sin(angle), readout_len)],
                    "sine": [(-np.cos(angle), readout_len)],
                },
            }
        )



"""Class to interface with the SPI Rack Qutech Delft."""
from qblox_instruments import SpiRack
from qibo.config import log, raise_error

from qibolab.instruments.abstract import Instrument, InstrumentException


class SPI(Instrument):
    property_wrapper = lambda parent, device, *parameter: property(
        lambda self: device.get(parameter[0]),
        lambda self, x: parent._set_device_parameter(device, *parameter, value=x),
    )

    def __init__(self, name, address):
        super().__init__(name, address)
        self.device: SpiRack = None
        self.s4g_modules_settings = {}
        self.d5a_modules_settings = {}
        self.dacs = {}
        self.device_parameters = {}

    def connect(self):
        """Connects to the instrument using the IP address set in the runcard."""
        if not self.is_connected:
            for attempt in range(3):
                try:
                    self.device = SpiRack(self.name, self.address)
                    self.is_connected = True
                    break
                except KeyError as exc:
                    log.info(f"Unable to connect:\n{str(exc)}\nRetrying...")
                    self.name += "_" + str(attempt)
                except Exception as exc:
                    log.info(f"Unable to connect:\n{str(exc)}\nRetrying...")
            if not self.is_connected:
                raise InstrumentException(self, f"Unable to connect to {self.name}")
        else:
            raise_error(Exception, "There is an open connection to the instrument already")

    def _set_device_parameter(self, target, *parameters, value):
        if self.is_connected:
            key = target.name + "." + parameters[0]
            if not key in self.device_parameters:
                for parameter in parameters:
                    if not hasattr(target, parameter):
                        raise Exception(f"The instrument {self.name} does not have parameters {parameter}")
                    target.set(parameter, value)
                self.device_parameters[key] = value
            elif self.device_parameters[key] != value:
                for parameter in parameters:
                    target.set(parameter, value)
                self.device_parameters[key] = value
        else:
            raise Exception("There is no connection to the instrument {self.name}")

    def setup(self, **kwargs):
        # Init S4g and D5a modules in SPI mapped on runcard
        if self.is_connected:
            # TODO: Check data format from yml
            #      Make d5g modules optional in runcard
            #      Define span values in setup
            #      Implement parameters cache
            #      export current / voltage properties (and make them sweepable)
            if "s4g_modules" in kwargs:
                self.s4g_modules_settings = kwargs["s4g_modules"]
            if "d5a_modules" in kwargs:
                self.d5a_modules_settings = kwargs["d5a_modules"]

            for channel, settings in self.s4g_modules_settings.items():
                module_number = settings[0]
                port_number = settings[1]
                module_name = f"S4g_module{module_number}"
                current = settings[2]
                if not module_name in self.device.instrument_modules:
                    self.device.add_spi_module(settings[0], "S4g", module_name)
                device = self.device.instrument_modules[module_name].instrument_modules["dac" + str(port_number - 1)]
                self.dacs[channel] = type(
                    "S4g_dac",
                    (),
                    {
                        "current": self.property_wrapper(device, "current"),
                        "device": device,
                    },
                )()
                self.dacs[channel].device.span("range_min_bi")
                # self.dacs[channel].current = current

            for channel, settings in self.d5a_modules_settings.items():
                module_number = settings[0]
                port_number = settings[1]
                module_name = f"D5a_module{module_number}"
                voltage = settings[2]
                if not module_name in self.device.instrument_modules:
                    self.device.add_spi_module(settings[0], "D5a", module_name)
                device = self.device.instrument_modules[module_name].instrument_modules["dac" + str(port_number - 1)]
                self.dacs[channel] = type(
                    "D5a_dac",
                    (),
                    {
                        "voltage": self.property_wrapper(device, "voltage"),
                        "device": device,
                    },
                )()
                self.dacs[channel].device.span("range_min_bi")
                # self.dacs[channel].voltage = voltage
        else:
            raise_error(Exception, "There is no connection to the instrument")

    def set_SPI_DACS_to_cero(self):
        self.device.set_dacs_zero()

    def get_SPI_IDN(self):
        return self.device.IDN()

    def get_SPI_temperature(self):
        return self.device.temperature()

    def get_SPI_battery_voltage(self):
        return self.device.battery_voltages()

    def disconnect(self):
        if self.is_connected:
            self.is_connected = False

    def close(self):
        if self.is_connected:
            self.device.close()
            self.is_connected = False

    def start(self):
        # Set the dacs to the values stored for each qubit in the runcard
        if self.is_connected:
            for channel, settings in self.s4g_modules_settings.items():
                current = settings[2]
                # Check current current of the module and warning
                if abs(self.dacs[channel].current) > 0.010:
                    log.info(
                        f"WARNING: S4g module {settings[0]} - port {settings[1]} current was: {self.dacs[channel].current}, now setting current to: {current}"
                    )
                self.dacs[channel].current = current

            for channel, settings in self.d5a_modules_settings.items():
                voltage = settings[2]
                # Check current current of the module and warning
                if abs(self.dacs[channel].voltage) > 0.010:
                    log.info(
                        f"WARNING: D5a module {settings[0]} - port {settings[1]} voltage was: {self.dacs[channel].voltage}, now setting voltage to: {voltage}"
                    )
                self.dacs[channel].voltage = voltage

    def stop(self):
        # if self.is_connected:
        #     self.device.set_dacs_zero()
        return



from dataclasses import dataclass, field
from typing import ClassVar, Dict, Optional

from qm import generate_qua_script, qua
from qm.qua import declare, for_
from qm.QuantumMachinesManager import QuantumMachinesManager

from qibolab import AveragingMode
from qibolab.instruments.abstract import Controller

from .config import IQPortId, QMConfig, QMPort
from .sequence import Sequence
from .sweepers import sweep


@dataclass
class QMOPX(Controller):
    """Instrument object for controlling Quantum Machines (QM) OPX controllers.

    Playing pulses on QM controllers requires a ``config`` dictionary and a program
    written in QUA language. The ``config`` file is generated in parts in the following places
    in the ``register_*`` methods. The controllers, elements and pulses are all
    registered after a pulse sequence is given, so that the config contains only
    elements related to the participating qubits.
    The QUA program for executing an arbitrary qibolab ``PulseSequence`` is written in
    ``play`` and ``play_pulses`` and executed in ``execute_program``.

    Args:
        name (str): Name of the instrument instance.
        address (str): IP address and port for connecting to the OPX instruments.
    """

    PortType: ClassVar = QMPort

    name: str
    address: str

    manager: Optional[QuantumMachinesManager] = None
    """Manager object used for controlling the QM OPXs."""
    config: QMConfig = field(default_factory=QMConfig)
    """Configuration dictionary required for pulse execution on the OPXs."""
    is_connected: bool = False
    """Boolean that shows whether we are connected to the QM manager."""
    time_of_flight: int = 0
    """Time of flight used for hardware signal integration."""
    smearing: int = 0
    """Smearing used for hardware signal integration."""
    _ports: Dict[IQPortId, QMPort] = field(default_factory=dict)
    """Dictionary holding the ports of controllers that are connected."""
    script_file_name: Optional[str] = "qua_script.txt"
    """Name of the file that the QUA program will dumped in that after every execution.
    If ``None`` the program will not be dumped.
    """

    def __post_init__(self):
        super().__init__(self.name, self.address)

    def connect(self):
        """Connect to the QM manager."""
        host, port = self.address.split(":")
        self.manager = QuantumMachinesManager(host, int(port))

    def setup(self):
        """Deprecated method."""
        # controllers are defined when registering pulses
        pass

    def start(self):
        # TODO: Start the OPX flux offsets?
        pass

    def stop(self):
        """Close all running Quantum Machines."""
        # TODO: Use logging
        # log.warn("Closing all Quantum Machines.")
        print("Closing all Quantum Machines.")
        self.manager.close_all_quantum_machines()

    def disconnect(self):
        """Disconnect from QM manager."""
        if self.is_connected:
            self.manager.close()
            self.is_connected = False

    def execute_program(self, program):
        """Executes an arbitrary program written in QUA language.

        Args:
            program: QUA program.

        Returns:
            TODO
        """
        machine = self.manager.open_qm(self.config.__dict__)
        return machine.execute(program)

    @staticmethod
    def fetch_results(result, ro_pulses):
        """Fetches results from an executed experiment.

        Defined as ``@staticmethod`` because it is overwritten
        in :class:`qibolab.instruments.qm.simulator.QMSim`.
        """
        # TODO: Update result asynchronously instead of waiting
        # for all values, in order to allow live plotting
        # using ``handles.is_processing()``
        handles = result.result_handles
        handles.wait_for_all_values()
        results = {}
        for qmpulse in ro_pulses:
            pulse = qmpulse.pulse
            results[pulse.qubit] = results[pulse.serial] = qmpulse.acquisition.fetch(handles)
        return results

    def play(self, qubits, sequence, options):
        return self.sweep(qubits, sequence, options)

    def sweep(self, qubits, sequence, options, *sweepers):
        if not sequence:
            return {}

        buffer_dims = [len(sweeper.values) for sweeper in reversed(sweepers)]
        if options.averaging_mode is AveragingMode.SINGLESHOT:
            buffer_dims.append(options.nshots)

        # register flux elements for all qubits so that they are
        # always at sweetspot even when they are not used
        for qubit in qubits.values():
            if qubit.flux:
                self.config.register_flux_element(qubit)

        qmsequence = Sequence.create(qubits, sequence, sweepers, self.config, self.time_of_flight, self.smearing)
        # play pulses using QUA
        with qua.program() as experiment:
            n = declare(int)
            for qmpulse in qmsequence.ro_pulses:
                threshold = qubits[qmpulse.pulse.qubit].threshold
                iq_angle = qubits[qmpulse.pulse.qubit].iq_angle
                qmpulse.declare_output(options, threshold, iq_angle)

            with for_(n, 0, n < options.nshots, n + 1):
                sweep(list(sweepers), qubits, qmsequence, options.relaxation_time, self.config)

            with qua.stream_processing():
                for qmpulse in qmsequence.ro_pulses:
                    qmpulse.acquisition.download(*buffer_dims)

        if self.script_file_name is not None:
            with open(self.script_file_name, "w") as file:
                file.write(generate_qua_script(experiment, self.config.__dict__))

        result = self.execute_program(experiment)
        return self.fetch_results(result, qmsequence.ro_pulses)

    def play_sequences(self, qubits, sequence, options):
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

import networkx as nx
import numpy as np
from qibo.backends import NumpyBackend
from qibo.config import log, raise_error
from qibo.models import Circuit


def find_gates_qubits_pairs(circuit: Circuit):
    """Translate qibo circuit into a list of pairs of qubits to be used by the transpiler.

    Args:
        circuit (qibo.models.Circuit): circuit to be transpiled.

    Returns:
        translated_circuit (list): list containing qubits targeted by two qubit gates
    """
    translated_circuit = []
    for gate in circuit.queue:
        if len(gate.qubits) == 2:
            translated_circuit.append(sorted(gate.qubits))
        if len(gate.qubits) >= 3:
            raise_error(ValueError, "Gates targeting more than 2 qubits are not supported")
    return translated_circuit


class Placer(ABC):
    """A placer implements the initial logical-physical qubit mapping"""

    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        self.connectivity = connectivity

    @abstractmethod
    def __call__(self, circuit: Circuit) -> dict:
        """Find initial qubit mapping

        Args:
            circuit (qibo.models.Circuit): circuit to be mapped.

        Returns:
            initial_layout (dict): dictionary containing the initial logical to physical qubit mapping.
        """


class Router(ABC):
    """A router implements the mapping of a circuit on a specific hardware"""

    @abstractmethod
    def __init__(self, connectivity: nx.Graph, *args):
        self.connectivity = connectivity

    @abstractmethod
    def __call__(self, circuit: Circuit, initial_layout: dict) -> Tuple[Circuit, dict]:
        """Match circuit to hardware connectivity

        Args:
            circuit (qibo.models.Circuit): circuit to be routed.
            initial_layout (dict): dictionary containing the initial logical to physical qubit mapping.

        Returns:
            matched_circuit (qibo.models.Circuit): routed circuit
            final_layout (dict): dictionary containing the final logical to physical qubit mapping.
        """


class Transpiler(ABC):
    """A transpiler is a transformation from a circuit to another circuit."""

    @abstractmethod
    def is_satisfied(self, circuit: Circuit) -> bool:
        """Checks if the circuit satisfies this transpiler's requirements.

        In some cases this is computationally easier to check than applying
        the transpilation transformations.
        """

    def check_execution(self, original_circuit: Circuit, transpiled_circuit: Optional[Circuit] = None):
        """Checks that the transpiled circuit agrees with the original using simulation.

        Args:
            original_circuit (:class:`qibo.models.Circuit`): Original circuit.
            transpiled_circuit (:class:`qibo.models.Circuit`): Transpiled circuit.
                If not given, it will be calculated using
                :meth:`qibolab.transpilers.abstract.Transpiler.transpile`.
        """
        if transpiled_circuit is None:
            transpiled_circuit, _ = self(original_circuit)

        backend = NumpyBackend()
        target_state = backend.execute_circuit(original_circuit).state()
        final_state = backend.execute_circuit(transpiled_circuit).state()
        fidelity = np.abs(np.dot(np.conj(target_state), final_state))
        np.testing.assert_allclose(fidelity, 1.0)
        log.info("Transpiler test passed.")

    @abstractmethod
    def __call__(self, circuit: Circuit) -> Tuple[Circuit, List[int]]:
        """Apply the transpiler transformation on a given circuit.

        Args:
            circuit (:class:`qibo.models.Circuit`): Circuit to transpile.

        Retruns:
            circuit (:class:`qibo.models.Circuit`): Circuit after transpilation.
            qubit_map (list): Order of qubits in the transpiled circuit.
        """



import itertools

import numpy as np
from qibo import __version__ as qibo_version
from qibo.backends import NumpyBackend
from qibo.config import raise_error
from qibo.states import CircuitResult

from qibolab import ExecutionParameters
from qibolab import __version__ as qibolab_version
from qibolab import create_platform
from qibolab.compilers import Compiler
from qibolab.platform import Platform
from qibolab.transpilers import Pipeline


class QibolabBackend(NumpyBackend):
    def __init__(self, platform, runcard=None):
        super().__init__()
        self.name = "qibolab"
        if isinstance(platform, Platform):
            self.platform = platform
        else:
            self.platform = create_platform(platform, runcard)
        self.versions = {
            "qibo": qibo_version,
            "numpy": self.np.__version__,
            "qibolab": qibolab_version,
        }
        self.compiler = Compiler.default()
        self.transpiler = Pipeline.default(self.platform.two_qubit_native_types)

    def apply_gate(self, gate, state, nqubits):  # pragma: no cover
        raise_error(NotImplementedError, "Qibolab cannot apply gates directly.")

    def apply_gate_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        raise_error(NotImplementedError, "Qibolab cannot apply gates directly.")

    def assign_measurements(self, measurement_map, circuit_result):
        """Assigning measurement outcomes to :class:`qibo.states.MeasurementResult` for each gate.

        This allows properly obtaining the measured shots from the :class:`qibo.states.CircuitResult`
        object returned by the circuit execution.

        Args:
            measurement_map (dict): Map from each measurement gate to the sequence of
                readout pulses implementing it.
            circuit_result (:class:`qibo.states.CircuitResult`): Circuit result object
                containing the readout measurement shots. This is created in ``execute_circuit``.
        """
        readout = circuit_result.execution_result
        for gate, sequence in measurement_map.items():
            _samples = (readout[pulse.serial].samples for pulse in sequence.pulses)
            samples = list(filter(lambda x: x is not None, _samples))
            gate.result.backend = self
            gate.result.register_samples(np.array(samples).T)

    def execute_circuit(
        self, circuit, initial_state=None, nshots=None, fuse_one_qubit=False, check_transpiled=False
    ):  # pragma: no cover
        """Executes a quantum circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to execute.
            initial_state (:class:`qibo.models.circuit.Circuit`): Circuit to prepare the initial state.
                If ``None`` the default ``|00...0>`` state is used.
            nshots (int): Number of shots to sample from the experiment.
                If ``None`` the default value provided as hardware_avg in the
                calibration yml will be used.
            fuse_one_qubit (bool): If ``True`` it fuses one qubit gates during
                transpilation to reduce circuit depth.
            check_transpiled (bool): If ``True`` it checks that the transpiled
                circuit is equivalent to the original using simulation.

        Returns:
            CircuitResult object containing the results acquired from the execution.
        """
        if isinstance(initial_state, type(circuit)):
            return self.execute_circuit(
                circuit=initial_state + circuit,
                nshots=nshots,
                fuse_one_qubit=fuse_one_qubit,
                check_transpiled=check_transpiled,
            )
        if initial_state is not None:
            raise_error(
                ValueError,
                "Hardware backend only supports circuits as initial states.",
            )

        if self.transpiler is None or self.transpiler.is_satisfied(circuit):
            native_circuit = circuit
        else:
            # Transform a circuit into proper connectivity and native gates
            native_circuit, qubit_map = self.transpiler(circuit)
            # TODO: Use the qubit map to properly map measurements
            if check_transpiled:
                self.transpiler.check_execution(circuit, native_circuit)

        # Transpile the native circuit into a sequence of pulses ``PulseSequence``
        sequence, measurement_map = self.compiler.compile(native_circuit, self.platform)

        if not self.platform.is_connected:
            self.platform.connect()
            self.platform.setup()

        # Execute the pulse sequence on the platform
        self.platform.start()
        readout = self.platform.execute_pulse_sequence(
            sequence,
            ExecutionParameters(nshots=nshots),
        )
        self.platform.stop()
        result = CircuitResult(self, circuit, readout, nshots)
        self.assign_measurements(measurement_map, result)
        return result

    def circuit_result_tensor(self, result):
        raise_error(
            NotImplementedError,
            "Qibolab cannot return state vector in tensor representation.",
        )

    def circuit_result_representation(self, result: CircuitResult):
        # TODO: Consider changing this to a more readable format.
        # this must return a ``str`` because it is used in ``CircuitResult.__repr__``.
        return str(result.execution_result)

    def circuit_result_probabilities(self, result: CircuitResult, qubits=None):
        """Returns the probability of the qubit being in state ``|0>``."""
        if qubits is None:  # pragma: no cover
            qubits = [self.platform.get_qubit(q) for q in result.measurement_gate.qubits]

        # basic classification
        probabilities = []
        for qubit in qubits:
            # execution_result[qubit] provides the latest acquisition data for the corresponding qubit
            qubit_result = result.execution_result[qubit]
            if qubit_result.samples is None:
                mean_state0 = complex(self.platform.qubits[qubit].mean_gnd_states)
                mean_state1 = complex(self.platform.qubits[qubit].mean_exc_states)
                measurement = complex(qubit_result.I, qubit_result.Q)
                d0 = abs(measurement - mean_state0)
                d1 = abs(measurement - mean_state1)
                d01 = abs(mean_state0 - mean_state1)
                p = (d1**2 + d01**2 - d0**2) / 2 / d01**2
                probabilities.append([p, 1 - p])
            else:
                outcomes, counts = np.unique(qubit_result.samples, return_counts=True)
                probabilities.append([0, 0])
                for i, c in zip(outcomes.astype(int), counts):
                    probabilities[-1][i] = c / result.nshots

        # bring probabilities to the format returned by simulation
        return np.array(
            [
                np.prod([p[b] for p, b in zip(probabilities, bitstring)])
                for bitstring in itertools.product([0, 1], repeat=len(qubits))
            ]
        )



from bisect import bisect
from typing import List

import numpy as np
from qibo.config import raise_error

from qibolab.instruments.abstract import Instrument, InstrumentException
from qibolab.pulses import Pulse


class TektronixAWG5204(Instrument):
    def __init__(self, name, address):
        super().__init__(name, address)
        # Phase offset for each channel for IQ sideband optimziation
        self.channel_phase: "list[float]" = []
        # Time buffer at the start and end of the pulse sequence to ensure that the minimum samples of the instrument are reached
        self.pulse_buffer: float = 1e-6
        self.device = None
        self.sample_rate = None

    rw_property_wrapper = lambda parameter: property(
        lambda self: self.device.get(parameter),
        lambda self, x: self.device.set(parameter, x),
    )
    rw_property_wrapper("sample_rate")

    def connect(self):
        if not self.is_connected:
            from qcodes.instrument_drivers.tektronix.AWG70000A import AWG70000A

            try:
                self.device = AWG70000A(self.name, self.address, num_channels=4)
            except Exception as exc:
                raise InstrumentException(self, str(exc))
            self.is_connected = True
        else:
            raise_error(Exception, "There is an open connection to the instrument already")

    def setup(self, **kwargs):
        if self.is_connected:
            # Set AWG to external reference, 10 MHz
            self.device.write("CLOC:SOUR EFIX")
            # Set external trigger to 1V
            self.device.write("TRIG:LEV 1")
            self.sample_rate = kwargs.pop("sample_rate")

            resolution = kwargs.pop("resolution")
            amplitude = kwargs.pop("amplitude")
            offset = kwargs.pop("offset")

            for idx, channel in enumerate(range(1, self.device.num_channels + 1)):
                awg_ch = getattr(self.device, f"ch{channel}")
                awg_ch.awg_amplitude(amplitude[idx])
                awg_ch.resolution(resolution)
                self.device.write(f"SOURCE{channel}:VOLTAGE:LEVEL:IMMEDIATE:OFFSET {offset[idx]}")

            self.__dict__.update(kwargs)
        else:
            raise_error(Exception, "There is no connection to the instrument")

    def generate_waveforms_from_pulse(self, pulse: Pulse, time_array: np.ndarray):
        """Generates a numpy array based on the pulse parameters

        Arguments:
            pulse (qibolab.pulses.Pulse | qibolab.pulses.ReadoutPulse): Pulse to be compiled
            time_array (numpy.ndarray): Array corresponding to the global time
        """
        i_ch, q_ch = pulse.channel

        i = pulse.envelope_i * np.cos(2 * np.pi * pulse.frequency * time_array + pulse.phase + self.channel_phase[i_ch])
        q = (
            -1
            * pulse.envelope_i
            * np.sin(2 * np.pi * pulse.frequency * time_array + pulse.phase + self.channel_phase[q_ch])
        )
        return i, q

    def translate(self, sequence: List[Pulse], nshots=None):
        """
        Translates the pulse sequence into a numpy array.

        Arguments:
            sequence (qibolab.pulses.Pulse[]): Array containing pulses to be fired on this instrument.
            nshots (int): Number of repetitions.
        """

        # First create np arrays for each channel
        start = min(pulse.start for pulse in sequence)
        end = max(pulse.start + pulse.duration for pulse in sequence)
        time_array = np.arange(
            start * 1e-9 - self.pulse_buffer,
            end * 1e-9 + self.pulse_buffer,
            1 / self.sample_rate,
        )
        waveform_arrays = np.zeros((self.device.num_channels, len(time_array)))

        for pulse in sequence:
            start_index = bisect(time_array, pulse.start * 1e-9)
            end_index = bisect(time_array, (pulse.start + pulse.duration) * 1e-9)
            i_ch, q_ch = pulse.channel
            i, q = self.generate_waveforms_from_pulse(pulse, time_array[start_index:end_index])
            waveform_arrays[i_ch, start_index:end_index] += i
            waveform_arrays[q_ch, start_index:end_index] += q

        return waveform_arrays

    def upload(self, waveform: np.ndarray):
        """Uploads a nchannels X nsamples array to the AWG, load it into memory and assign it to the channels for playback."""

        # TODO: Add additional check to ensure all waveforms are of the same size? Should be caught by qcodes driver anyway.
        if len(waveform) != self.device.num_channels:
            raise_error(Exception, "Invalid waveform given")

        # Clear existing waveforms in memory
        self.device.write("WLIS:WAV:DEL ALL")

        # Upload waveform, load into memory and assign to each channel
        for idx, channel in enumerate(range(1, self.device.num_channels + 1)):
            awg_ch = getattr(self.device, f"ch{channel}")
            wfmx = self.device.makeWFMXFile(waveform[idx], awg_ch.awg_amplitude())
            self.device.sendWFMXFile(wfmx, f"ch{channel}.wfmx")
            self.device.loadWFMXFile(f"ch{channel}.wfmx")
            self.device.write(f'SOURce{channel}:CASSet:WAVeform "ch{channel}"')

    def start(self):
        pass

    def play_sequence(self):
        for channel in range(1, self.device.num_channels + 1):
            awg_ch = getattr(self.device, f"ch{channel}")
            awg_ch.state(1)
            self.device.write(f"SOURce{channel}:RMODe TRIGgered")
            self.device.write(f"SOURce{channel}:TINPut ATR")
        self.device.play()
        self.device.wait_for_operation_to_complete()

    def stop(self):
        self.device.stop()
        for channel in range(1, self.device.num_channels + 1):
            awg_ch = getattr(self.device, f"ch{channel}")
            awg_ch.state(0)
        self.device.wait_for_operation_to_complete()

    def disconnect(self):
        if self.is_connected:
            self.device.stop()
            self.device.close()
            self.is_connected = False

    def __del__(self):
        self.disconnect()

    def close(self):
        if self.is_connected:
            self.stop()
            self.device.close()
            self.is_connected = False


class MCAttenuator(Instrument):
    """Driver for the MiniCircuit RCDAT-8000-30 variable attenuator."""

    def connect(self):
        pass

    def setup(self, attenuation: float, **kwargs):
        """Assigns the attenuation level on the attenuator.

        Arguments:
            attenuation(float
            ): Attenuation setting in dB. Ranges from 0 to 35.

        """
        import urllib3

        http = urllib3.PoolManager()
        http.request("GET", f"http://{self.address}/SETATT={attenuation}")

    def start(self):
        pass

    def stop(self):
        pass

    def disconnect(self):
        pass


class QuicSyn(Instrument):
    """Driver for the National Instrument QuicSyn Lite local oscillator."""

    def connect(self):
        if not self.is_connected:
            import pyvisa as visa

            rm = visa.ResourceManager()
            try:
                self.device = rm.open_resource(self.address)
            except Exception as exc:
                raise InstrumentException(self, str(exc))
            self.is_connected = True

    def setup(self, frequency: float, **kwargs):
        """Sets the frequency in Hz."""
        if self.is_connected:
            self.device.write("0601")
            self.frequency(frequency)

    def frequency(self, frequency):
        self.device.write("FREQ {:f}Hz".format(frequency))

    def start(self):
        """Starts the instrument."""
        self.device.write("0F01")

    def stop(self):
        """Stops the instrument."""
        self.device.write("0F00")

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        if self.is_connected:
            self.stop()
            self.device.close()
            self.is_connected = False


class AlazarADC(Instrument):
    """Driver for the AlazarTech ATS9371 ADC."""

    def __init__(self, name, address):
        super().__init__(name, address)
        self.controller = None
        self.device = None

    def connect(self):
        if not self.is_connected:
            from qcodes.instrument_drivers.AlazarTech.AlazarADC import (  # pylint: disable=E0401, E0611
                ADCController,
            )
            from qcodes.instrument_drivers.AlazarTech.ATS9371 import (  # pylint: disable=E0401, E0611
                AlazarTech_ATS9371,
            )

            try:
                self.device = AlazarTech_ATS9371(self.address)
                self.controller = ADCController(self.name, self.address)
            except Exception as exc:
                raise InstrumentException(self, str(exc))
            self.is_connected = True

    def setup(self, trigger_volts, **kwargs):
        """Sets the frequency in Hz."""
        if self.is_connected:
            input_range_volts = 2.5
            trigger_level_code = int(128 + 127 * trigger_volts / input_range_volts)
            with self.device.syncing():
                self.device.clock_source("EXTERNAL_CLOCK_10MHz_REF")
                self.device.external_sample_rate(1_000_000_000)
                self.device.clock_edge("CLOCK_EDGE_RISING")
                self.device.decimation(1)
                self.device.coupling1("DC")
                self.device.coupling2("DC")
                self.device.channel_range1(0.02)
                self.device.channel_range2(0.02)
                self.device.impedance1(50)
                self.device.impedance2(50)
                self.device.bwlimit1("DISABLED")
                self.device.bwlimit2("DISABLED")
                self.device.trigger_operation("TRIG_ENGINE_OP_J")
                self.device.trigger_engine1("TRIG_ENGINE_J")
                self.device.trigger_source1("EXTERNAL")
                self.device.trigger_slope1("TRIG_SLOPE_POSITIVE")
                self.device.trigger_level1(trigger_level_code)
                self.device.trigger_engine2("TRIG_ENGINE_K")
                self.device.trigger_source2("DISABLE")
                self.device.trigger_slope2("TRIG_SLOPE_POSITIVE")
                self.device.trigger_level2(128)
                self.device.external_trigger_coupling("DC")
                self.device.external_trigger_range("ETR_2V5")
                self.device.trigger_delay(0)
                self.device.timeout_ticks(0)

            samples = kwargs.pop("samples")
            self.controller.samples = samples
            self.__dict__.update(kwargs)

    def arm(self, nshots, readout_start):
        with self.device.syncing():
            self.device.trigger_delay(int(int((readout_start * 1e-9 + 4e-6) / 1e-9 / 8) * 8))
        self.controller.arm(nshots)

    def play_sequence_and_acquire(self):
        """
        this method performs an acquisition, which is the get_cmd for the
        acquisiion parameter of this instrument
        :return:
        """
        raw = self.device.acquire(acquisition_controller=self.controller, **self.controller.acquisitionkwargs)
        return self.process_result(raw)

    def process_result(self, readout_frequency=100e6, readout_channels=[0, 1]):
        """Returns the processed signal result from the ADC.

        Arguments:
            readout_frequency (float): Frequency to be used for signal processing.
            readout_channels (int[]): Channels to be used for signal processing.

        Returns:
            ampl (float): Amplitude of the processed signal.
            phase (float): Phase shift of the processed signal in degrees.
            it (float): I component of the processed signal.
            qt (float): Q component of the processed signal.
        """

        input_vec_I = self.device._processed_data[readout_channels[0]]
        input_vec_Q = self.device._processed_data[readout_channels[1]]
        it = 0
        qt = 0
        for i in range(self.device.samples_per_record):
            it += input_vec_I[i] * np.cos(2 * np.pi * readout_frequency * self.device.time_array[i])
            qt += input_vec_Q[i] * np.cos(2 * np.pi * readout_frequency * self.device.time_array[i])
        phase = np.arctan2(qt, it)
        ampl = np.sqrt(it**2 + qt**2)

        return ampl, phase, it, qt

    def start(self):
        """Starts the instrument."""

    def stop(self):
        """Stops the instrument."""

    def disconnect(self):
        if self.is_connected:
            self.device.close()
            self.controller.close()
